# -*- encoding: utf-8 -*-
"""
keri.kli.commands.migrate.list module

"""
import argparse
import logging

from keri import help
from hio.base import doing
from prettytable import PrettyTable

from keri.app.cli.common import existing

logger = help.ogler.getLogger()

def handler(args):
    """
    List local LMDB database migrations and their completion status

    Args:
        args(Namespace): arguments object from command line
    """
    lister = ListDoer(args)
    return [lister]


parser = argparse.ArgumentParser(description='Lists the local LMDB migrations and their completion status')
parser.set_defaults(handler=handler,
                    transferable=True)

# Parameters for basic structure of database
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--temp', '-t', help='create a temporary keystore, used for testing', default=False)

# Parameters for Manager creation
# passcode => bran
parser.add_argument('--passcode', '-p', help='21 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)


class ListDoer(doing.Doer):

    def __init__(self, args):
        self.args = args
        super(ListDoer, self).__init__()

    def recur(self, tyme):
        tab = PrettyTable()
        tab.field_names = ["Num", "Name", "Date Completed"]
        tab.align["Name"] = "l"

        hby = existing.setupHby(name=self.args.name, base=self.args.base,
                                bran=self.args.bran if self.args.bran else None, temp=self.args.temp)

        for idx, (name, dater) in enumerate(hby.db.complete()):
            print(name, dater)
            date = dater.datetime.strftime("%Y-%m-%d %H:%M") if dater is not None else "Not Run"
            tab.add_row((f"{idx + 1}", f"{name}", date))

        print(tab)
        return True



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse
import datetime
import json
import sys

from keri import help
from hio.base import doing

from keri import kering
from keri.app import indirecting
from keri.app.cli.common import existing, terming
from keri.core import scheming
from keri.help import helping
from keri.vdr import credentialing, verifying

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='List credentials and check mailboxes for any newly issued credentials')
parser.set_defaults(handler=lambda args: list_credentials(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--alias', '-a', help='human readable alias for the identifier to whom the credential was issued',
                    default=None)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran

parser.add_argument("--verbose", "-V", help="print full JSON of all credentials", action="store_true")
parser.add_argument("--poll", "-P", help="Poll mailboxes for any issued credentials", action="store_true")
parser.add_argument("--issued", "-i", help="Display credentials that this AID has issued.",
                    action="store_true")
parser.add_argument("--said", "-s", help="Display only the SAID of found credentials, one per line.",
                    action="store_true")
parser.add_argument("--schema", help="Display only credentials with the given schema SAID.",
                    action="store", default=None)


def list_credentials(args):
    """ Command line list credential registries handler

    """
    ld = ListDoer(name=args.name,
                  alias=args.alias,
                  base=args.base,
                  bran=args.bran,
                  verbose=args.verbose,
                  poll=args.poll,
                  said=args.said,
                  issued=args.issued,
                  schema=args.schema)
    return [ld]


class ListDoer(doing.DoDoer):

    def __init__(self, name, alias, base, bran, verbose=False, poll=False, said=False, issued=False, schema=None):
        self.verbose = verbose
        self.poll = poll
        self.said = said
        self.issued = issued
        self.schema = schema

        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        if alias is None:
            alias = existing.aliasInput(self.hby)

        self.hab = self.hby.habByName(alias)
        self.rgy = credentialing.Regery(hby=self.hby, name=name, base=base)
        self.vry = verifying.Verifier(hby=self.hby, reger=self.rgy.reger)
        self.mbx = indirecting.MailboxDirector(hby=self.hby, topics=['/credential'], verifier=self.vry)

        doers = [self.mbx, doing.doify(self.listDo)]

        super(ListDoer, self).__init__(doers=doers)

    def listDo(self, tymth, tock=0.0):
        """ Check for any credential messages in mailboxes and list all held credentials

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        Returns:  doifiable Doist compatible generator method

        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        if self.poll:
            end = helping.nowUTC() + datetime.timedelta(seconds=5)
            sys.stdout.write(f"Checking mailboxes for any {'issued' if self.issued else 'received'} credentials")
            sys.stdout.flush()
            while helping.nowUTC() < end:
                sys.stdout.write(".")
                sys.stdout.flush()
                if "/credential" in self.mbx.times:
                    end = self.mbx.times['/credential'] + datetime.timedelta(seconds=5)
                yield 1.0
            print("\n")

        if self.issued:
            saids = self.rgy.reger.issus.get(keys=self.hab.pre)
        else:
            saids = self.rgy.reger.subjs.get(keys=self.hab.pre)

        if self.schema is not None:
            scads = self.rgy.reger.schms.get(keys=self.schema)
            saids = [saider for saider in saids if saider.qb64 in [saider.qb64 for saider in scads]]

        if self.said:
            for said in saids:
                print(said.qb64)
        else:
            print(f"Current {'issued' if self.issued else 'received'}"
                  f" credentials for {self.hab.name} ({self.hab.pre}):\n")
            creds = self.rgy.reger.cloneCreds(saids, self.hab.db)
            for idx, cred in enumerate(creds):
                sad = cred['sad']
                status = cred["status"]
                schema = sad['s']
                scraw = self.mbx.verifier.resolver.resolve(schema)
                if not scraw:
                    raise kering.ConfigurationError("Credential schema {} not found".format(schema))

                schemer = scheming.Schemer(raw=scraw)
                print(f"Credential #{idx+1}: {sad['d']}")
                print(f"    Type: {schemer.sed['title']}")
                if status['et'] == 'iss' or status['et'] == 'bis':
                    print(f"    Status: Issued {terming.Colors.OKGREEN}{terming.Symbols.CHECKMARK}{terming.Colors.ENDC}")
                elif status['et'] == 'rev' or status['et'] == 'brv':
                    print(f"    Status: Revoked {terming.Colors.FAIL}{terming.Symbols.FAILED}{terming.Colors.ENDC}")
                else:
                    print(f"    Status: Unknown")
                print(f"    Issued by {sad['i']}")
                print(f"    Issued on {status['dt']}")

                if self.verbose:
                    bsad = json.dumps(sad, indent=2)
                    print("    Full Credential:")
                    for line in bsad.splitlines():
                        print(f"\t{line}")

        self.remove([self.mbx])



# -*- encoding: utf-8 -*-
"""
KERI
keri.help package

utility functions

"""
# Setup module global ogler as package logger factory. This must be done on
#  import to ensure global is defined so all modules in package have access to
#  logggers via ogling.ogler.getLoggers(). May always change level and reopen log file
#  if need be

from hio.help import ogling

import logging

# Custom TRACE log level configuration
TRACE = 5              # TRACE (5) logging level value between DEBUG (10) and NOTSET (0)
logging.TRACE = TRACE  # add TRACE logging level to logging module
logging.addLevelName(logging.TRACE, "TRACE")
def trace(self, message, *args, **kwargs):
    """Trace logging function - logs message if TRACE (5) level enabled"""
    if self.isEnabledFor(TRACE):
        self._log(TRACE, message, args, **kwargs)
logging.Logger.trace = trace

#  want help.ogler always defined by default
ogler = ogling.initOgler(prefix='keri', syslogged=False)  # inits once only on first import

from .helping import nowIso8601, toIso8601, fromIso8601



# -*- encoding: utf-8 -*-
"""
KERI
keri.vdr.verifying module

VC verifier support
"""
import datetime
import logging
from typing import Type

from hio.help import decking

from .. import help, kering
from ..core import parsing, coring, scheming
from ..help import helping
from ..vdr import eventing
from ..vdr.viring import Reger

logger = help.ogler.getLogger()


class Verifier:
    """
    Verifier class accepts and validates TEL events.

    """
    TimeoutPSE = 3600  # seconds to timeout partially signed credential escrow
    TimeoutMRE = 3600  # seconds to timeout missing registry escrows
    TimeoutMRI = 3600  # seconds to timeout missing issuer escrows
    TimeoutBCE = 3600  # seconds to timeout missing issuer escrows

    def __init__(self, hby, reger=None, creds=None, cues=None, expiry=36000000000):
        """
        Initialize Verifier instance

        Parameters:
            hby (Habery): for this verifier's context
            reger (Reger): database instance
            creds (decking.Deck): inbound credentials for handler
            cues (decking.Deck): outbound cue messages from handler

        """
        self.hby = hby
        self.reger = reger if reger is not None else Reger(name=self.hby.name, temp=self.hby.temp)
        self.creds = creds if creds is not None else decking.Deck()  # subclass of deque
        self.cues = cues if cues is not None else decking.Deck()  # subclass of deque
        self.CredentialExpiry = expiry

        self.inited = False
        self.tvy = None
        self.psr = None
        self.resolver = None

        if self.hby.inited:
            self.setup()

    def setup(self):
        """ Delayed initialization of instance by createing .tvy and .psr.

        Should not be called until .hab is initialized

        """
        self.tvy = eventing.Tevery(reger=self.reger, db=self.hby.db, local=False)
        self.psr = parsing.Parser(framed=True, kvy=self.hby.kvy, tvy=self.tvy)
        self.resolver = scheming.CacheResolver(db=self.hby.db)

        self.inited = True

    @property
    def tevers(self):
        """ Returns .db.tevers
        """
        return self.reger.tevers

    def processMessages(self, creds=None):
        """ Process message dicts in msgs or if msgs is None in .msgs

        Parameters:
            creds (decking.Deck): each entry is dict that matches call signature of
                .processCredential
        """
        if creds is None:
            creds = self.creds

        while creds:
            self.processCredential(**creds.pull())

    def processCredential(self, creder, prefixer, seqner, saider):
        """ Credential data and signature(s) verification

        Verify the data of the credential against the schema, the SAID of the credential and
        the CESR Proof on the credential and if valid, store the credential

        Parameters:
            creder (Creder): that contains the credential to process
            prefixer (Prefixer): prefix of source anchoring KEL or TEL event
            seqner (Seqner): sequence number of source anchoring KEL or TEL event
            saider (Saider): SAID of source anchoring KEL or TEL event

        """
        regk = creder.regi
        vcid = creder.said
        schema = creder.schema
        prov = creder.edge if creder.edge is not None else {}

        if regk not in self.tevers:  # registry event not found yet
            if self.escrowMRE(creder, prefixer, seqner, saider):
                self.cues.append(dict(kin="telquery", q=dict(ri=regk, i=vcid, issr=creder.issuer)))
            raise kering.MissingRegistryError("registry identifier {} not in Tevers".format(regk))

        state = self.tevers[regk].vcState(vcid)
        if state is None:  # credential issuance event not found yet
            if self.escrowMRE(creder, prefixer, seqner, saider):
                self.cues.append(dict(kin="telquery", q=dict(ri=regk, i=vcid)))
            raise kering.MissingRegistryError("credential identifier {} not in Tevers".format(vcid))

        dtnow = helping.nowUTC()
        dte = helping.fromIso8601(state.dt)
        if (dtnow - dte) > datetime.timedelta(seconds=self.CredentialExpiry):
            if self.escrowMRE(creder, prefixer, seqner, saider):
                self.cues.append(dict(kin="telquery", q=dict(ri=regk, i=vcid)))
            raise kering.MissingRegistryError("credential identifier {} is out of date".format(vcid))
        elif state.et in (coring.Ilks.rev, coring.Ilks.brv):  # no escrow, credential has been revoked
            logger.error("credential {} in registrying is not in issued state".format(vcid, regk))
            # Log this and continue instead of the previous exception so we save a revoked credential.
            # raise kering.InvalidCredentialStateError("..."))

        # Verify the credential against the schema
        scraw = self.resolver.resolve(schema)
        if not scraw:
            if self.escrowMSE(creder, prefixer, seqner, saider):
                self.cues.append(dict(kin="query", q=dict(r="schema", said=schema)))
            raise kering.MissingSchemaError("schema {} not in cache".format(schema))

        schemer = scheming.Schemer(raw=scraw)
        try:
            schemer.verify(creder.raw)
        except kering.ValidationError as ex:
            print("Credential {} is not valid against schema {}: {}"
                  .format(creder.said, schema, ex))
            raise kering.FailedSchemaValidationError("Credential {} is not valid against schema {}: {}"
                                                     .format(creder.said, schema, ex))

        if isinstance(prov, list):
            edges = prov
        elif isinstance(prov, dict):
            edges = [prov]
        else:
            print(f"Invalid type for edges: {prov}")
            raise kering.ValidationError(f"invalid type for edges: {prov}")

        for edge in edges:
            for label, node in edge.items():
                if label in ('d', 'o'):  # SAID or Operator of this edge block
                    continue
                nodeSaid = node["n"]
                op = node['o'] if 'o' in node else None
                state = self.verifyChain(nodeSaid, op, creder.issuer)
                if state is None:
                    self.escrowMCE(creder, prefixer, seqner, saider)
                    self.cues.append(dict(kin="proof",  said=nodeSaid))
                    raise kering.MissingChainError("Failure to verify credential {} chain {}({})"
                                                   .format(creder.said, label, nodeSaid))

                dtnow = helping.nowUTC()
                dte = helping.fromIso8601(state.dt)
                if (dtnow - dte) > datetime.timedelta(seconds=self.CredentialExpiry):
                    self.escrowMCE(creder, prefixer, seqner, saider)
                    self.cues.append(dict(kin="query", q=dict(r="tels", pre=nodeSaid)))
                    raise kering.MissingChainError("Failure to verify credential {} chain {}({})"
                                                   .format(creder.said, label, nodeSaid))
                elif state.et in (coring.Ilks.rev, coring.Ilks.brv):
                    raise kering.RevokedChainError("Failure to verify credential {} chain {}({})"
                                                   .format(creder.said, label, nodeSaid))
                else:  # VcStatus == VcStates.Issued
                    logger.info("Successfully validated credential chain {} for credential {}"
                                .format(label, creder.said))

        self.saveCredential(creder, prefixer, seqner, saider)
        self.cues.append(dict(kin="saved", creder=creder))

    def escrowMRE(self, creder, prefixer, seqner, saider):
        """ Missing Registry Escrow

        Parameters:
            creder (Creder): that contains the credential to process
            prefixer (Prefixer): prefix (AID or TEL) of event anchoring credential
            seqner (Seqner): sequence number of event anchoring credential
            saider (Diger) digest of anchoring event for credential

        """
        key = creder.said

        self.reger.logCred(creder, prefixer, seqner, saider)
        return self.reger.mre.put(keys=key, val=coring.Dater())

    def escrowMCE(self, creder, prefixer, seqner, saider):
        """ Missing Chain Escrow

        Parameters:
            creder (Creder): that contains the credential to process
            prefixer (Prefixer): prefix (AID or TEL) of event anchoring credential
            seqner (Seqner): sequence number of event anchoring credential
            saider (Diger) digest of anchoring event for credential

        """
        key = creder.said

        self.reger.logCred(creder, prefixer, seqner, saider)
        return self.reger.mce.put(keys=key, val=coring.Dater())

    def escrowMSE(self, creder, prefixer, seqner, saider):
        """
        Missing Credential Schema Escrow


        Parameters:
            creder (Creder): that contains the credential to process
            prefixer (Prefixer): prefix (AID or TEL) of event anchoring credential
            seqner (Seqner): sequence number of event anchoring credential
            saider (Diger) digest of anchoring event for credential

        """
        key = creder.said

        self.reger.logCred(creder, prefixer, seqner, saider)
        return self.reger.mse.put(keys=key, val=coring.Dater())

    def processEscrows(self):
        """ Process all escrows once each

        """

        self._processEscrow(self.reger.mce, self.TimeoutMRI, kering.MissingChainError)
        self._processEscrow(self.reger.mse, self.TimeoutMRI, kering.MissingSchemaError)
        self._processEscrow(self.reger.mre, self.TimeoutMRE, kering.MissingRegistryError)

    def _processEscrow(self, db, timeout, etype: Type[Exception]):
        """ Generic credential escrow processing

        Parameters:
            db (LMDBer): escrow database table to process
            timeout (float): escrow specific message timeout
            etype (TypeOf(Exception)): exception class to catch and ignore

        """
        for (said,), dater in db.getItemIter():
            creder, prefixer, seqner, saider = self.reger.cloneCred(said)

            try:

                dtnow = helping.nowUTC()
                dte = helping.fromIso8601(dater.dts)
                if (dtnow - dte) > datetime.timedelta(seconds=timeout):
                    # escrow stale so raise ValidationError which unescrows below
                    logger.info("Verifier unescrow error: Stale event escrow "
                                " at said = %s\n", said)

                    raise kering.ValidationError("Stale event escrow "
                                                 "at said = {}.".format(said))

                self.processCredential(creder, prefixer, seqner, saider)

            except etype as ex:
                if logger.isEnabledFor(logging.TRACE):
                    logger.trace("Verifier unescrow failed: %s\n", ex.args[0])
                    logger.exception("Verifier unescrow failed: %s\n", ex.args[0])
            except Exception as ex:  # log diagnostics errors etc
                # error other than missing sigs so remove from PA escrow
                db.rem(said)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Verifier unescrowed: %s\n", ex.args[0])
                else:
                    logger.error("Verifier unescrowed: %s\n", ex.args[0])
            else:
                db.rem(said)
                logger.info("Verifier unescrow succeeded in valid group op: "
                            "creder = %s", creder.said)
                logger.debug("Creder body=\n%s\n", creder.pretty())

    def saveCredential(self, creder, prefixer, seqner, saider):
        """ Write the credential and associated indicies to the database

        Parameters:
            creder (Creder): that contains the credential to process
            prefixer (Prefixer): prefix (AID or TEL) of event anchoring credential
            seqner (Seqner): sequence number of event anchoring credential
            saider (Diger) digest of anchoring event for credential

        """
        self.reger.logCred(creder, prefixer, seqner, saider)

        schema = creder.schema.encode("utf-8")
        issuer = creder.issuer.encode("utf-8")

        # Look up indicies
        saider = coring.Saider(qb64=creder.said)
        self.reger.saved.pin(keys=saider.qb64b, val=saider)
        self.reger.issus.add(keys=issuer, val=saider)
        self.reger.schms.add(keys=schema, val=saider)

        if 'i' in creder.attrib:
            subject = creder.attrib["i"].encode("utf-8")
            self.reger.subjs.add(keys=subject, val=saider)

    def query(self, pre, regk, vcid, *, dt=None, dta=None, dtb=None, **kwa):
        """ Returns query message for querying registry
        """

        serder = eventing.query(regk=regk, vcid=vcid, dt=dt, dta=dta,
                                dtb=dtb, **kwa)
        hab = self.hby.habs[pre]
        return hab.endorse(serder, last=True)

    def verifyChain(self, nodeSaid, op, issuer):
        """ Verifies the node credential at the end of an edge

        Parameters:
            nodeSaid: (str): qb64 SAID of node credential
            op(str): edge operator
            issuer (str) qb64 AID of issuer

        Returns:
            Serder: transaction event state notification message

        """
        said = self.reger.saved.get(keys=nodeSaid)
        if said is None:
            return None

        creder = self.reger.creds.get(keys=nodeSaid)

        if op not in ['I2I', 'DI2I', 'NI2I']:
            op = 'I2I' if 'i' in creder.attrib else 'NI2I'

        if op != 'NI2I':
            if 'i' not in creder.attrib:
                return None

            iss = self.reger.subjs.get(keys=creder.attrib['i'])
            if iss is None:
                return None

            if op == 'I2I' and issuer != creder.attrib['i']:
                return None

            if op == "DI2I":
                raise NotImplementedError()

        if creder.regi not in self.tevers:
            return None

        tever = self.tevers[creder.regi]

        state = tever.vcState(nodeSaid)
        if state is None:
            return None

        return state



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse
import datetime

import sys
from keri import help
from hio.base import doing

from keri.app import indirecting
from keri.app.cli.common import displaying, existing
from keri.core import serdering
from keri.help import helping

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Initialize a prefix')
parser.set_defaults(handler=lambda args: handler(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--prefix', help='qb64 identifier prefix to display', required=True)
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument("--poll", "-P", help="Poll mailboxes for any events", action="store_true")

parser.add_argument("--verbose", "-V", help="print JSON of all current events", action="store_true")


def handler(args):
    kwa = dict(args=args)
    kever = KeverDoer(name=args.name, base=args.base, bran=args.bran, prefix=args.prefix, poll=args.poll,
                      verbose=args.verbose)
    return [kever]


class KeverDoer(doing.DoDoer):

    def __init__(self, name, base, bran, prefix, poll=False, verbose=False):
        self.prefix = prefix
        self.poll = poll
        self.verbose = verbose
        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.mbx = indirecting.MailboxDirector(hby=self.hby,
                                               topics=["/receipt", "/replay", "/multisig", "/credential", "/delegate",
                                                       "/challenge", "/oobi"])
        doers = [self.mbx, doing.doify(self.kevers)]
        super(KeverDoer, self).__init__(doers=doers)

    def kevers(self, tymth, tock=0.0, **opts):
        """ Command line status handler

        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield tock)

        if self.poll:
            end = helping.nowUTC() + datetime.timedelta(seconds=5)
            sys.stdout.write(f"Checking mailboxes for any events")
            sys.stdout.flush()
            while helping.nowUTC() < end:
                sys.stdout.write(".")
                sys.stdout.flush()
                yield 1.0
            print("\n")

        if self.prefix not in self.hby.kevers:
            print(f"identifier prefix {self.prefix} is not known locally")
        else:
            displaying.printExternal(self.hby, self.prefix)

            if self.verbose:
                kever = self.hby.kevers[self.prefix]
                print("\nWitnesses:\t")
                for idx, wit in enumerate(kever.wits):
                    print(f'\t{idx + 1}. {wit}')
                print()

                cloner = self.hby.db.clonePreIter(pre=self.prefix, fn=0)  # create iterator at 0
                for msg in cloner:
                    srdr = serdering.SerderKERI(raw=msg)
                    print(srdr.pretty())
                    print()

        self.remove([self.mbx])

        return True



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse
import sys

from keri import help
from hio.base import doing

from keri.app.cli.common import existing
from keri.core import serdering

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='List credentials and check mailboxes for any newly issued credentials')
parser.set_defaults(handler=lambda args: export(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--alias', '-a', help='human readable alias for the identifier to whom the credential was issued',
                    required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument("--files", help="export artifacts to individual files keyed off of AIDs or SAIDS, default is "
                                    "stdout", action="store_true")
parser.add_argument("--ends", help="export service end points", action="store_true")


def export(args):
    """ Command line list credential registries handler

    """

    ed = ExportDoer(name=args.name,
                    alias=args.alias,
                    base=args.base,
                    bran=args.bran,
                    ends=args.ends,
                    files=args.files,)
    return [ed]


class ExportDoer(doing.DoDoer):

    def __init__(self, name, alias, base, bran, ends, files):
        self.files = files
        self.ends = ends

        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.hab = self.hby.habByName(alias)

        doers = [doing.doify(self.exportDo)]

        super(ExportDoer, self).__init__(doers=doers)

    def exportDo(self, tymth, tock=0.0):
        """ Export credential from store and any related material

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        Returns:  doifiable Doist compatible generator method

        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        self.output(said=self.hab.pre)

    def output(self, said):
        self.outputKEL(pre=said)

        if self.ends:
            self.outputEnds(pre=said)

        sys.stdout.flush()

    def outputKEL(self, pre):
        f = None
        if self.files:
            f = open(f"{pre}-kel.cesr", "w")

        for msg in self.hby.db.clonePreIter(pre=pre):
            if f is not None:
                f.write(msg.decode("utf-8"))
            else:
                serder = serdering.SerderKERI(raw=msg)
                atc = msg[serder.size:]
                sys.stdout.write(serder.raw.decode("utf-8"))
                sys.stdout.write(atc.decode("utf-8"))

        if f is not None:
            f.close()

    def outputEnds(self, pre):
        f = None
        if self.files:
            f = open(f"{pre}-ends.cesr", "w")

        msgs = self.hab.replyToOobi(aid=pre, role="controller")
        for msg in msgs:
            if f is not None:
                f.write(msg.decode("utf-8"))
            else:
                serder = serdering.SerderKERI(raw=msg)
                atc = msg[serder.size:]
                sys.stdout.write(serder.raw.decode("utf-8"))
                sys.stdout.write(atc.decode("utf-8"))



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse

from keri import help
from hio.base import doing

from keri.app.cli.common import existing

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='List existing identifiers')
parser.set_defaults(handler=lambda args: list_identifiers(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran

parser.add_argument("--verbose", "-V", help="print JSON of all current events", action="store_true")


def list_identifiers(args):
    """ Command line list handler

    """
    kwa = dict(args=args)
    return [doing.doify(ids, **kwa)]


def ids(tymth, tock=0.0, **opts):
    _ = (yield tock)

    args = opts["args"]
    name = args.name
    base = args.base
    bran = args.bran

    with existing.existingHby(name=name, base=base, bran=bran) as hby:
        for hab in hby.habs.values():
            print(f"{hab.name} ({hab.pre})")



# -*- encoding: utf-8 -*-
"""
keri.kli.commands.oobi module

"""
import argparse
from urllib.parse import urlparse

import sys
from keri import help
from hio.base import doing

from keri import kering
from keri.app.cli.common import existing

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Generate and print role OOBIs for the AID of the provide alias.')
parser.set_defaults(handler=lambda args: handler(args),
                    transferable=True)

# Parameters for basic structure of database
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the new identifier prefix', default=None)
parser.add_argument("--role", "-r", help="role of oobis to generate", required=True)

# Parameters for Manager access
# passcode => bran
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)


def handler(args):
    """ command line method for generating oobies

    Parameters:
        args(Namespace): parse args namespace object

    """
    kwa = dict(args=args)
    return [doing.doify(generate, **kwa)]


def generate(tymth, tock=0.0, **opts):
    """ Command line status handler

    """
    _ = (yield tock)
    args = opts["args"]
    name = args.name
    alias = args.alias
    base = args.base
    bran = args.bran
    role = args.role

    with existing.existingHby(name=name, base=base, bran=bran) as hby:
        if alias is None:
            alias = existing.aliasInput(hby)

        hab = hby.habByName(name=alias)
        if role in (kering.Roles.witness,):
            if not hab.kever.wits:
                print(f"{alias} identifier {hab.pre} does not have any witnesses.")
                sys.exit(-1)

            for wit in hab.kever.wits:
                urls = hab.fetchUrls(eid=wit, scheme=kering.Schemes.http) or hab.fetchUrls(eid=wit, scheme=kering.Schemes.https)
                if not urls:
                    raise kering.ConfigurationError(f"unable to query witness {wit}, no http endpoint")
            
                url = urls[kering.Schemes.http] if kering.Schemes.http in urls else urls[kering.Schemes.https]
                up = urlparse(url)
                print(f"{up.scheme}://{up.hostname}:{up.port}/oobi/{hab.pre}/witness")
        elif role in (kering.Roles.controller,):
            urls = hab.fetchUrls(eid=hab.pre, scheme=kering.Schemes.http) or hab.fetchUrls(eid=hab.pre, scheme=kering.Schemes.https)
            if not urls:
                print(f"{alias} identifier {hab.pre} does not have any controller endpoints")
                return
            url = urls[kering.Schemes.http] if kering.Schemes.http in urls else urls[kering.Schemes.https]
            up = urlparse(url)
            print(f"{up.scheme}://{up.hostname}:{up.port}/oobi/{hab.pre}/controller")



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse
import json

from keri import help
from hio.base import doing
from mnemonic import mnemonic

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Generate a cryptographically random challenge phrase')
parser.set_defaults(handler=lambda args: handler(args),
                    transferable=True)
parser.add_argument('--strength', '-s', help='Cryptographic strength in bits.  Defaults to 128', default=128,
                    required=False)
parser.add_argument("--out", "-o", help="Output type [words|string|json] of phrase.  Default is json",
                    choices=["words", "string", "json"], default="json", required=False)


def handler(args):
    """
    Generate a cryptographically random challenge phrase

    Args:
        args(Namespace): arguments object from command line
    """
    kwa = dict(args=args)
    return [doing.doify(generate, **kwa)]


def generate(tock=0.0, **opts):
    _ = (yield tock)

    args = opts["args"]
    generateWords(args.strength, args.out)
    return True


def generateWords(strength, out):
    mnem = mnemonic.Mnemonic(language='english')
    s = strength
    strength = int(s) if s is not None else 128

    words = mnem.generate(strength=strength)
    if out == "json":
        msg = words.split(" ")
        print(json.dumps(msg))
    elif out == "string":
        print(words)
    else:
        for w in words.split(" "):
            print(w)

    return words.split(" ")



1.1# -*- encoding: utf-8 -*-
"""
keri.core.escrowing module

"""
import datetime
import logging
from typing import Type

from keri.core import coring
from keri import help
from keri import kering
from keri.core import eventing
from keri.db import subing
from keri.help import helping

logger = help.ogler.getLogger()


class Broker:

    def __init__(self, db, subkey, timeout=3600):
        self.db = db
        self.timeout = timeout

        # State support datetime stamps and signatures indexed and not-indexed
        # all ksn  kdts (key state datetime serializations) maps said to date-time
        self.daterdb = subing.CesrSuber(db=self.db, subkey=subkey + '-dts.', klas=coring.Dater)

        # all reply messages that holdkey state messages.
        # Maps replay messages that hold key state said to serialization. ksns are
        # versioned sads ( with version string) so use Serder to deserialize and
        # use  .kdts, .ksgs, and .kcgs for datetimes and signatures
        self.serderdb = subing.SerderSuber(db=self.db, subkey=subkey + '-sns.')

        # RegStateRecords used as basis for registry state notices in replies
        #self.rsrdb = koming.Komer(db=self.db,
                                   #schema=viring.RegStateRecord,
                                   #subkey=subkey + '-sns.')

        # all key state ksgs (ksn indexed signature serializations) maps ksn quadkeys
        # given by quadruple (saider.qb64, subkeyer.qb64, seqner.q64, diger.qb64)
        #  of reply and trans signer's key state est evt to val Siger for each
        # signature.
        self.tigerdb = subing.CesrIoSetSuber(db=self.db, subkey=subkey + '-sgs.', klas=coring.Siger)

        # all key state kcgs  (ksn non-indexed signature serializations) maps ksn SAID
        # to couple (Verfer, Cigar) of nontrans signer of signature in Cigar
        # nontrans qb64 of subkeyer is same as Verfer
        self.cigardb = subing.CatCesrIoSetSuber(db=self.db, subkey=subkey + '-cgs.',
                                                klas=(coring.Verfer, coring.Cigar))

        # all key state escrows indices of partially signed ksn messages. Maps
        # route in reply to single (Saider,)  of escrowed ksn.
        # Routes such as /ksn/{aid} or /tsn/registry/{aid}
        self.escrowdb = subing.CesrIoSetSuber(db=self.db, subkey=subkey + '-nes', klas=coring.Saider)

        # transaction state SAID database for successfully saved transaction state notices
        # maps key=(prefix, aid) to val=said of transaction state
        self.saiderdb = subing.CesrSuber(db=self.db, subkey=subkey + '-nas.', klas=coring.Saider)

    def current(self, keys):
        return self.saiderdb.get(keys=keys)

    def processEscrowState(self, typ, processReply, extype: Type[Exception]):
        """ Process escrows for reply messages

        Process escrows for reply messages. Escrows are keyed by reply pre
        and val is reply said

        triple (prefixer, seqner, diger)
        quadruple (prefixer, seqner, diger, siger)

        Parameters:
            typ (str): escrow type
            processReply (func): function to call to process each message taken out of escrow
            extype (Type[Exception]): the expected exception type if the message should remain in escrow

        """
        for (typ, pre, aid, ion), saider in self.escrowdb.getIoItemIter(keys=(typ,)):
            try:
                tsgs = eventing.fetchTsgs(db=self.tigerdb, saider=saider)

                keys = (saider.qb64,)
                dater = self.daterdb.get(keys=keys)
                serder = self.serderdb.get(keys=keys)
                vcigars = self.cigardb.get(keys=keys)

                try:
                    if not (dater and serder and (tsgs or vcigars)):
                        raise ValueError(f"Missing escrow artifacts at said={saider.qb64}"
                                         f"for pre={pre}.")

                    cigars = []
                    if vcigars:
                        for (verfer, cigar) in vcigars:
                            cigar.verfer = verfer
                            cigars.append(cigar)

                    # do date math for stale escrow
                    if ((helping.nowUTC() - dater.datetime) >
                            datetime.timedelta(seconds=self.timeout)):
                        # escrow stale so raise ValidationError which unescrows below
                        msg = f"{typ} escrow unescrow error: Stale txn state escrow at pre = {pre}"
                        logger.trace(msg)
                        raise kering.ValidationError(msg)

                    processReply(serder=serder, saider=saider, route=serder.ked["r"],
                                 cigars=cigars, tsgs=tsgs, aid=aid)

                except extype as ex:
                    # still waiting on missing prior event to validate
                    if logger.isEnabledFor(logging.TRACE):
                        logger.trace("%s escrow unescrow attempt failed: %s\n", typ, ex.args[0])
                        logger.exception("%s escrow  unescrow attempt failed: %s\n", typ, ex.args[0])

                except Exception as ex:  # other error so remove from reply escrow
                    self.escrowdb.remIokey(iokeys=(typ, pre, aid, ion))  # remove escrow
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.exception("%s escrow other error on unescrow: %s\n", typ, ex.args[0])
                    else:
                        logger.error("%s escrow other error on unescrow: %s\n", typ, ex.args[0])

                else:  # unescrow succeded
                    self.escrowdb.remIokey(iokeys=(typ, pre, aid, ion))  # remove escrow only
                    logger.info("%s escrow unescrow succeeded for txn state = %s",
                                typ, serder.said)
                    logger.debug("TXN State Body=\n%s\n", serder.pretty())

            except Exception as ex:  # log diagnostics errors etc
                self.escrowdb.remIokey(iokeys=(typ, pre, aid, ion))  # remove escrow
                self.removeState(saider)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("%s escrow unescrowed due to error: %s\n", typ, ex.args[0])
                else:
                    logger.error("%s escrow unescrowed due to error: %s\n", typ, ex.args[0])

    def escrowStateNotice(self, *, typ, pre, aid, serder, saider, dater, cigars=None, tsgs=None):
        """
        Escrow reply by route

        Parameters:
            typ (str): escrow type
            pre (str): identifier of key state
            aid (str): identifier of authorizer of key state
            serder (Serder): instance of reply msg (SAD)
            saider (Saider): instance  from said in serder (SAD)
            dater (Dater): instance from date-time in serder (SAD)
            cigars (list): of Cigar instances that contain nontrans signing couple
                          signature in .raw and public key in .verfer

            tsgs (Iterable): of quadruples of form (prefixer, seqner, diger, siger) where:
                prefixer is pre of trans endorser
                seqner is sequence number of trans endorser's est evt for keys for sigs
                diger is digest of trans endorser's est evt for keys for sigs
                siger is indexed sig from trans endorser's key from est evt
        """
        cigars = cigars if cigars is not None else []
        tsgs = tsgs if tsgs is not None else []

        keys = (saider.qb64,)
        self.daterdb.put(keys=keys, val=dater)  # first one idempotent
        self.serderdb.put(keys=keys, val=serder)  # first one idempotent

        for prefixer, seqner, diger, sigers in tsgs:  # iterate over each tsg
            quadkeys = (saider.qb64, prefixer.qb64, f"{seqner.sn:032x}", diger.qb64)
            self.tigerdb.put(keys=quadkeys, vals=sigers)
        for cigar in cigars:  # process each couple to verify sig and write to db
            self.cigardb.put(keys=keys, vals=[(cigar.verfer, cigar)])

        return self.escrowdb.put(keys=(typ, pre, aid), vals=[saider])  # overwrite

    def updateReply(self, aid, serder, saider, dater):
        """
        Update Reply SAD in database given by by serder and associated databases
        for attached cig couple or sig quadruple.
        Overwrites val at key if already exists.

        Parameters:
            aid (str): identifier of key state
            serder (Serder): instance of reply msg (SAD)
            saider (Saider): instance  from said in serder (SAD)
            dater (Dater): instance from date-time in serder (SAD)
        """
        keys = (saider.qb64,)

        # Add source of ksn to the key for DATEs too...  (source AID, ksn AID)
        self.daterdb.put(keys=keys, val=dater)  # first one idempotent
        self.serderdb.pin(keys=keys, val=serder)  # first one idempotent
        # Add source of ksn to the key...  (source AID, ksn AID)
        self.saiderdb.pin(keys=(serder.sad["a"]["i"], aid), val=saider)  # overwrite

    def removeState(self, saider):
        if saider:
            keys = (saider.qb64,)

            self.tigerdb.trim(keys=(saider.qb64, ""))  # remove whole branch
            self.cigardb.rem(keys=keys)
            self.serderdb.rem(keys=keys)
            self.daterdb.rem(keys=keys)



# -*- encoding: utf-8 -*-
"""
KERI
keri.db.koming module

"""
import types
import json
from dataclasses import dataclass
from typing import Type, Union
from collections.abc import Iterable

import cbor2
import msgpack
import lmdb


from . import dbing
from .. import help
from ..core import coring
from ..help import helping

logger = help.ogler.getLogger()



class KomerBase:
    """
    KomerBase is a base class for Komer (Keyspace Object Mapper) subclasses that
    each use a dataclass as the object mapped via serialization to an dber LMDB
    database subclass.
    Each Komer .schema is a dataclass class reference that is used to define
    the fields in each database entry. The base class is not meant to be instantiated.
    Use an instance of one of the subclasses instead.

    Attributes:
        db (dbing.LMDBer): instance of LMDB database manager class
        sdb (lmdb._Database): instance of named sub db lmdb for this Komer
        schema (Type[dataclass]): class reference of dataclass subclass
        kind (str): serialization/deserialization type from coring.Serials
        serializer (types.MethodType): serializer method
        deserializer (types.MethodType): deserializer method
        sep (str): separator for combining keys tuple of strs into key bytes
    """
    Sep = '.'  # separator for combining key iterables

    def __init__(self, db: dbing.LMDBer, *,
                 subkey: str = 'docs.',
                 schema: Type[dataclass],  # class not instance
                 kind: str = coring.Serials.json,
                 dupsort: bool = False,
                 sep: str = None,
                 **kwa):
        """
        Parameters:
            db (dbing.LMDBer): base db
            schema (Type[dataclass]):  reference to Class definition for dataclass sub class
            subkey (str):  LMDB sub database key
            kind (str): serialization/deserialization type
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
        """
        super(KomerBase, self).__init__()
        self.db = db
        self.sdb = self.db.env.open_db(key=subkey.encode("utf-8"), dupsort=dupsort)
        self.schema = schema
        self.kind = kind
        self.serializer = self._serializer(kind)
        self.deserializer = self._deserializer(kind)
        self.sep = sep if sep is not None else self.Sep


    def _tokey(self, keys: Union[str, bytes, memoryview, Iterable]):
        """
        Converts key to key str with proper separators and returns key bytes.
        If key is already str then returns. Else If key is iterable (non-str)
        of strs then joins with separator converts to bytes and returns

        Parameters:
           keys (Union[str, bytes, Iterable]): str, bytes, or Iterable of str.

        """
        if isinstance(keys, memoryview):  # memoryview of bytes
            return bytes(keys)  # return bytes
        if hasattr(keys, "encode"):  # str
            return keys.encode("utf-8")  # convert to bytes
        elif hasattr(keys, "decode"): # bytes
            return keys  # return as is
        return (self.sep.join(keys).encode("utf-8"))  # iterable so join


    def _tokeys(self, key: Union[str, bytes, memoryview]):
        """
        Converts key bytes to keys tuple of strs by decoding and then splitting
        at separator.

        Returns:
           keys (iterable): of str

        Parameters:
           key (Union[str, bytes]): str or bytes.

        """
        if isinstance(key, memoryview):  # memoryview of bytes
            key = bytes(key)
        return tuple(key.decode("utf-8").split(self.sep))


    def getItemIter(self, keys: Union[str, Iterable]=b""):
        """
        Returns:
            items (Iterator): of (key, val) tuples  over the all the items in
            subdb whose key startswith key made from keys. Keys may be keyspace
            prefix to return branches of key space. When keys is empty then
            returns all items in subdb

        Parameters:
            keys (Iterator): tuple of bytes or strs that may be a truncation of
                a full keys tuple in  in order to get all the items from
                multiple branches of the key space. If keys is empty then gets
                all items in database.

        """
        for key, val in self.db.getTopItemIter(db=self.sdb, key=self._tokey(keys)):
            yield (self._tokeys(key), self.deserializer(val))


    def _serializer(self, kind):
        """
        Parameters:
            kind (str): serialization
        """
        if kind == coring.Serials.mgpk:
            return self.__serializeMGPK
        elif kind == coring.Serials.cbor:
            return self.__serializeCBOR
        else:
            return self.__serializeJSON


    def _deserializer(self, kind):
        """
        Parameters:
            kind (str): deserialization
        """
        if kind == coring.Serials.mgpk:
            return self.__deserializeMGPK
        elif kind == coring.Serials.cbor:
            return self.__deserializeCBOR
        else:
            return self.__deserializeJSON


    def __deserializeJSON(self, val):
        if val is not None:
            val = helping.datify(self.schema, json.loads(bytes(val).decode("utf-8")))
            if not isinstance(val, self.schema):
                raise ValueError("Invalid schema type={} of value={}, expected {}."
                                 "".format(type(val), val, self.schema))
        return val


    def __deserializeMGPK(self, val):
        if val is not None:
            val = helping.datify(self.schema, msgpack.loads(bytes(val)))
            if not isinstance(val, self.schema):
                raise ValueError("Invalid schema type={} of value={}, expected {}."
                                 "".format(type(val), val, self.schema))
        return val


    def __deserializeCBOR(self, val):
        if val is not None:
            val = helping.datify(self.schema, cbor2.loads(bytes(val)))
            if not isinstance(val, self.schema):
                raise ValueError("Invalid schema type={} of value={}, expected {}."
                                 "".format(type(val), val, self.schema))
        return val


    def __serializeJSON(self, val):
        if val is not None:
            if not isinstance(val, self.schema):
                raise ValueError("Invalid schema type={} of value={}, expected {}."
                                 "".format(type(val), val, self.schema))
            val = json.dumps(helping.dictify(val),
                          separators=(",", ":"),
                          ensure_ascii=False).encode("utf-8")
        return val


    def __serializeMGPK(self, val):
        if val is not None:
            if not isinstance(val, self.schema):
                raise ValueError("Invalid schema type={} of value={}, expected {}."
                                 "".format(type(val), val, self.schema))
            val = msgpack.dumps(helping.dictify(val))
        return val


    def __serializeCBOR(self, val):
        if val is not None:
            if not isinstance(val, self.schema):
                raise ValueError("Invalid schema type={} of value={}, expected {}."
                                 "".format(type(val), val, self.schema))
            val = cbor2.dumps(helping.dictify(val))
        return val


class Komer(KomerBase):
    """
    Keyspace Object Mapper factory class.
    """
    def __init__(self,
                 db: dbing.LMDBer, *,
                 subkey: str = 'docs.',
                 schema: Type[dataclass],  # class not instance
                 kind: str = coring.Serials.json,
                 **kwa):
        """
        Parameters:
            db (dbing.LMDBer): base db
            schema (Type[dataclass]):  reference to Class definition for dataclass sub class
            subkey (str):  LMDB sub database key
            kind (str): serialization/deserialization type
        """
        super(Komer, self).__init__(db=db, subkey=subkey, schema=schema,
                                    kind=kind, dupsort=False, **kwa)

    def put(self, keys: Union[str, Iterable], val: dataclass):
        """
        Puts val at key made from keys. Does not overwrite

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (dataclass): instance of dataclass of type self.schema as value

        Returns:
            result (bool): True If successful, False otherwise, such as key
                              already in database.
        """
        return (self.db.putVal(db=self.sdb,
                               key=self._tokey(keys),
                               val=self.serializer(val)))

    def pin(self, keys: Union[str, Iterable], val: dataclass):
        """
        Pins (sets) val at key made from keys. Overwrites.

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (dataclass): instance of dataclass of type self.schema as value

        Returns:
            result (bool): True If successful. False otherwise.
        """
        return (self.db.setVal(db=self.sdb,
                               key=self._tokey(keys),
                               val=self.serializer(val)))

    def get(self, keys: Union[str, Iterable]):
        """
        Gets val at keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
            val (dataclass):
            None if no entry at keys

        Usage:
            Use walrus operator to catch and raise missing entry
            if (val := mydb.get(keys)) is None:
                raise ExceptionHere
            use val here
        """
        return (self.deserializer(self.db.getVal(db=self.sdb,
                                  key=self._tokey(keys))))

    def getDict(self, keys: Union[str, Iterable]):
        """
        Gets dictified val at keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
            val (dict):
            None if no entry at keys

        Usage:
            Use walrus operator to catch and raise missing entry
            if (val := mydb.get(keys)) is None:
                raise ExceptionHere
            use val here
        """
        val = self.get(keys)
        return helping.dictify(val) if val is not None else None


    def rem(self, keys: Union[str, Iterable]):
        """
        Removes entry at keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
           result (bool): True if key exists so delete successful. False otherwise
        """
        return (self.db.delVal(db=self.sdb, key=self._tokey(keys)))


    def trim(self, keys: Union[str, Iterable]=b""):
        """
        Removes all entries whose keys startswith keys. Enables removal of whole
        branches of db key space. To ensure that proper separation of a branch
        include empty string as last key in keys. For example ("a","") deletes
        'a.1'and 'a.2' but not 'ab'

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
           result (bool): True if key exists so delete successful. False otherwise
        """
        return(self.db.delTopVal(db=self.sdb, key=self._tokey(keys)))


    def cntAll(self):
        """
        Return iterator over the all the items in subdb

        Returns:
            iterator: of tuples of keys tuple and val dataclass instance for
            each entry in db. Raises StopIteration when done

        Example:
            if key in database is "a.b" and val is serialization of dataclass
               with attributes x and y then returns
               (("a","b"), dataclass(x=1,y=2))
        """
        return self.db.cnt(db=self.sdb)


class IoSetKomer(KomerBase):
    """
    Insertion Ordered Set Keyspace Object Mapper factory class that supports
    a set of distinct entries at a given effective database key but with
    dupsort==False. Effective data model is that there are multiple values in a
    set of values where every member of the set has the same key (duplicate key).
    The set of values is an ordered set using insertion order. Any given value
    may appear only once in the set (not a list).

    This works similarly to the IO value duplicates for the LMDBer class with a
    sub db  of LMDB (dupsort==True) but without its size limitation of 511 bytes
    for each value when dupsort==True.
    Here the key is augmented with a hidden numbered suffix that provides a
    an ordered set of values at each effective key (duplicate key). The suffix
    is appended and stripped transparently. The set of multiple items with
    duplicate keys are retrieved in insertion order when iterating or as a list
    of the set elements.

    Attributes:
        db (dbing.LMDBer): instance of LMDB database manager class
        sdb (lmdb._Database): instance of named sub db lmdb for this Komer
        schema (Type[dataclass]): class reference of dataclass subclass
        kind (str): serialization/deserialization type from coring.Serials
        serializer (types.MethodType): serializer method
        deserializer (types.MethodType): deserializer method
        sep (str): separator for combining keys tuple of strs into key bytes
    """
    def __init__(self,
             db: dbing.LMDBer, *,
             subkey: str = 'recs.',
             schema: Type[dataclass],  # class not instance
             kind: str = coring.Serials.json,
             **kwa):
        """
        Parameters:
            db (dbing.LMDBer): base db
            schema (Type[dataclass]):  reference to Class definition for dataclass sub class
            subkey (str):  LMDB sub database key
            kind (str): serialization/deserialization type
        """
        super(IoSetKomer, self).__init__(db=db, subkey=subkey, schema=schema,
                                       kind=kind, dupsort=False, **kwa)


    def put(self, keys: Union[str, Iterable], vals: list):
        """
        Puts all vals at key made from keys. Does not overwrite. Puts all vals
        at effective key made from keys and hidden ordinal suffix.
        that are not already in set of vals at key. Does not overwrite.

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            vals (list): dataclass instances each of type self.schema as values

        Returns:
            result (bool): True If successful, False otherwise.

        Apparently always returns True (how .put works with dupsort=True)

        """
        vals = [self.serializer(val) for val in vals]
        return (self.db.putIoSetVals(db=self.sdb,
                                     key=self._tokey(keys),
                                     vals=vals,
                                     sep=self.sep))


    def add(self, keys: Union[str, Iterable], val: dataclass):
        """
        Add val to vals at effective key made from keys and hidden ordinal suffix.
        that is not already in set of vals at key. Does not overwrite.

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (dataclass): instance of type self.schema

        Returns:
            result (bool): True means unique value among duplications,
                              False means duplicte of same value already exists.

        """
        return (self.db.addIoSetVal(db=self.sdb,
                                    key=self._tokey(keys),
                                    val=self.serializer(val),
                                    sep=self.sep))


    def pin(self, keys: Union[str, Iterable], vals: list):
        """
        Pins (sets) vals at effective key made from keys and hidden ordinal suffix.
        Overwrites. Removes all pre-existing vals that share same effective keys
        and replaces them with vals

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            vals (list): dataclass instances each of type self.schema as values

        Returns:
            result (bool): True If successful, False otherwise.

        """
        key = self._tokey(keys)
        self.db.delIoSetVals(db=self.sdb, key=key)  # delete all values
        vals = [self.serializer(val) for val in vals]
        return (self.db.setIoSetVals(db=self.sdb,
                                     key=key,
                                     vals=vals,
                                     sep=self.sep))


    def get(self, keys: Union[str, Iterable]):
        """
        Gets dup vals list at key made from keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
            vals (list):  each item in list is instance of type self.schema
                          empty list if no entry at keys

        """
        return [self.deserializer(val) for val in
                    self.db.getIoSetValsIter(db=self.sdb,
                                             key=self._tokey(keys),
                                             sep=self.sep)]


    def getLast(self, keys: Union[str, Iterable]):
        """
        Gets last effective dup val at effective dup key made from keys

        Parameters:
            keys (tuple): of key strs to be combined to form effective key

        Returns:
            val (Type[dataclass]):  instance of type self.schema
                          None if no entry at keys

        """
        val = self.db.getIoSetValLast(db=self.sdb, key=self._tokey(keys))
        if val is not None:
            val = self.deserializer(val)
        return val


    def getIter(self, keys: Union[str, Iterable]):
        """
        Gets vals iterator at effecive key made from keys and hidden ordinal suffix.
        All vals in set of vals that share same effecive key are retrieved in
        insertion order.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            vals (Iterator):  str values. Raises StopIteration when done

        """
        for val in self.db.getIoSetValsIter(db=self.sdb,
                                            key=self._tokey(keys),
                                            sep=self.sep):
            yield self.deserializer(val)



    def cnt(self, keys: Union[str, Iterable]):
        """
        Return count of effective dup values at key made from keys, zero otherwise

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
        """
        return (self.db.cntIoSetVals(db=self.sdb,
                                     key=self._tokey(keys),
                                     sep=self.sep))


    def rem(self, keys: Union[str, Iterable], val=None):
        """
        Removes entry at keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (dataclass):  instance of effective dup val at key to delete
                              if val is None then remove all values at key

        Returns:
           result (bool): True if key exists so delete successful. False otherwise

        """
        if val is not None:
            val = self.serializer(val)
            return self.db.delIoSetVal(db=self.sdb,
                                       key=self._tokey(keys),
                                       val=val,
                                       sep=self.sep)
        else:
            return self.db.delIoSetVals(db=self.sdb,
                                       key=self._tokey(keys),
                                       sep=self.sep)


    def getItemIter(self, keys: Union[str, Iterable]=b""):
        """Get items iterator
        Returns:
            items (Iterator): of (key, val) tuples over the all the items in
            subdb whose effective key startswith key made from keys.
            Keys may be keyspace prefix in order to return branches of key space.
            When keys is empty then returns all items in subdb.
            Returned key in each item has ordinal suffix removed.

        Parameters:
            keys (Iterator): tuple of bytes or strs that may be a truncation of
                a full keys tuple in  in order to get all the items from
                multiple branches of the key space. If keys is empty then gets
                all items in database. Append "" to end of keys Iterable to
                ensure get properly separated top branch key.

        """
        for iokey, val in self.db.getTopItemIter(db=self.sdb, key=self._tokey(keys)):
            key, ion = dbing.unsuffix(iokey, sep=self.sep)
            yield (self._tokeys(key), self.deserializer(val))


    def getIoSetItem(self, keys: Union[str, Iterable]):
        """
        Gets (iokeys, val) ioitems list at key made from keys where key is
        apparent effective key and ioitems all have same apparent effective key

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            ioitems (Iterable):  each item in list is tuple (iokeys, val) where each
                    iokeys is actual key tuple including hidden suffix and
                    each val is str
                    empty list if no entry at keys


        """
        return ([(self._tokeys(iokey), self.deserializer(val)) for iokey, val
                    in self.db.getIoSetItems(db=self.sdb,
                                             key=self._tokey(keys),
                                             sep=self.sep)])


    def getIoSetItemIter(self, keys: Union[str, Iterable]):
        """
        Gets (iokeys, val) ioitems  iterator at key made from keys where key is
        apparent effective key and items all have same apparent effective key

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            ioitems (Iterator):  each item iterated is tuple (iokeys, val) where
                each iokeys is actual keys tuple including hidden suffix and
                each val is str
                empty list if no entry at keys.
                Raises StopIteration when done

        """
        for iokey, val in self.db.getIoSetItemsIter(db=self.sdb,
                                                    key=self._tokey(keys),
                                                    sep=self.sep):
            yield (self._tokeys(iokey), self.deserializer(val))


    def getIoItemIter(self, keys: Union[str, Iterable]=b""):
        """
        Returns:
            items (Iterator): tuple (key, val) over the all the items in
            subdb whose key startswith effective key made from keys.
            Keys may be keyspace prefix to return branches of key space.
            When keys is empty then returns all items in subdb.


        Parameters:
            keys (Iterable): tuple of bytes or strs that may be a truncation of
                a full keys tuple in  in order to get all the items from
                multiple branches of the key space. If keys is empty then gets
                all items in database. Append "" to end of keys Iterable to
                ensure get properly separated top branch key.

        """
        for iokey, val in self.db.getTopItemIter(db=self.sdb, key=self._tokey(keys)):
            yield (self._tokeys(iokey), self.deserializer(val))


    def remIokey(self, iokeys: Union[str, bytes, memoryview, Iterable]):
        """
        Removes entry at iokeys

        Parameters:
            iokeys (tuple): of key str or tuple of key strs to be combined in
                            order to form key

        Returns:
           result (bool): True if key exists so delete successful. False otherwise

        """
        return self.db.delIoSetIokey(db=self.sdb, iokey=self._tokey(iokeys))



class DupKomer(KomerBase):
    """
    Duplicate Keyspace Object Mapper factory class that supports multiple entries
    a given database key (lmdb dupsort == True).

    Do not use if Komer schema instance serialized is greater than 511 bytes.
    This is a limitation of dupsort==True sub dbs in LMDB
    """
    def __init__(self,
             db: dbing.LMDBer, *,
             subkey: str = 'recs.',
             schema: Type[dataclass],  # class not instance
             kind: str = coring.Serials.json,
             **kwa):
        """
        Parameters:
            db (dbing.LMDBer): base db
            schema (Type[dataclass]):  reference to Class definition for dataclass sub class
            subkey (str):  LMDB sub database key
            kind (str): serialization/deserialization type
        """
        super(DupKomer, self).__init__(db=db, subkey=subkey, schema=schema,
                                       kind=kind, dupsort=True, **kwa)



    def put(self, keys: Union[str, Iterable], vals: list):
        """
        Puts all vals at key made from keys. Does not overwrite. Adds to existing
        dup values at key if any. Duplicate means another entry at the same key
        but the entry is still a unique value. Duplicates are inserted in
        lexocographic order not insertion order. Lmdb does not insert a duplicate
        unless it is a unique value for that key.

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            vals (list): dataclass instances each of type self.schema as values

        Returns:
            result (bool): True If successful, False otherwise.

        Apparently always returns True (how .put works with dupsort=True)

        """
        vals = [self.serializer(val) for val in vals]
        return (self.db.putVals(db=self.sdb,
                                key=self._tokey(keys),
                                vals=vals))


    def add(self, keys: Union[str, Iterable], val: dataclass):
        """
        Add val to vals at key made from keys. Does not overwrite. Adds to existing
        dup values at key if any. Duplicate means another entry at the same key
        but the entry is still a unique value. Duplicates are inserted in
        lexocographic order not insertion order. Lmdb does not insert a duplicate
        unless it is a unique value for that key.

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (dataclass): instance of type self.schema

        Returns:
            result (bool): True means unique value among duplications,
                              False means duplicte of same value already exists.

        """
        return (self.db.addVal(db=self.sdb,
                               key=self._tokey(keys),
                               val=self.serializer(val)))


    def pin(self, keys: Union[str, Iterable], vals: list):
        """
        Pins (sets) vals at key made from keys. Overwrites. Removes all
        pre-existing dup vals and replaces them with vals

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            vals (list): dataclass instances each of type self.schema as values

        Returns:
            result (bool): True If successful, False otherwise.

        """
        key = self._tokey(keys)
        self.db.delVals(db=self.sdb, key=key)  # delete all values
        vals = [self.serializer(val) for val in vals]
        return (self.db.putVals(db=self.sdb,
                                key=key,
                                vals=vals))


    def get(self, keys: Union[str, Iterable]):
        """
        Gets dup vals list at key made from keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
            vals (list):  each item in list is instance of type self.schema
                          empty list if no entry at keys

        """
        return ([self.deserializer(val) for val in
                self.db.getValsIter(db=self.sdb, key=self._tokey(keys))])


    def getLast(self, keys: Union[str, Iterable]):
        """
        Gets last dup val at key made from keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
            val (Type[dataclass]):  instance of type self.schema
                          None if no entry at keys

        """
        val = self.db.getValLast(db=self.sdb, key=self._tokey(keys))
        if val is not None:
            val = self.deserializer(val)
        return val


    def getIter(self, keys: Union[str, Iterable]):
        """
        Gets dup vals iterator at key made from keys

        Duplicates are retrieved in lexocographic order not insertion order.

        Parameters:
            keys (tuple): of key strs to be combined in order to form key

        Returns:
            iterator:  vals each of type self.schema. Raises StopIteration when done

        """
        for val in self.db.getValsIter(db=self.sdb, key=self._tokey(keys)):
            yield self.deserializer(val)


    def cnt(self, keys: Union[str, Iterable]):
        """
        Return count of dup values at key made from keys, zero otherwise

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
        """
        return (self.db.cntVals(db=self.sdb, key=self._tokey(keys)))


    def rem(self, keys: Union[str, Iterable], val=None):
        """
        Removes entry at keys

        Parameters:
            keys (tuple): of key strs to be combined in order to form key
            val (dataclass):  instance of dup val at key to delete
                              if val is None then remove all values at key

        Returns:
           result (bool): True if key exists so delete successful. False otherwise

        """
        if val is not None:
            val = self.serializer(val)
        else:
            val = b''
        return (self.db.delVals(db=self.sdb, key=self._tokey(keys), val=val))




import argparse

from keri import help
from hio.base import doing

from keri.app import indirecting, habbing, grouping, forwarding
from keri.app.cli.common import existing
from keri.app.habbing import GroupHab
from keri.app.notifying import Notifier
from keri.core import serdering
from keri.core.eventing import SealEvent
from keri.peer import exchanging
from keri.vdr import credentialing

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Initialize a prefix')
parser.set_defaults(handler=lambda args: registryIncept(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.add_argument('--registry-name', '-r', help='Human readable name for registry, defaults to name of Habitat',
                    default=None)
parser.add_argument('--nonce', help='Unique random value to seed the credential registry',
                    default=None, required=False)
parser.add_argument("--no-backers", "-nb", help="do not allow setting up backers different from the ahcnoring KEL "
                                                "witnesses", default=True, action="store")
parser.add_argument('--backers', help='New set of backers different from the anchoring KEL witnesses.  Can '
                                      'appear multiple times', metavar="<prefix>", default=[], action="append",
                    required=False)
parser.add_argument("--establishment-only", "-eo", help="Only allow establishment events for the anchoring events of "
                                                        "this registry", default=False, action="store")
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the new identifier prefix', required=True)
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument('--usage', '-u', help='For multisig issuers, a message to other participants about how this'
                                          ' registry is to be used',
                    default=None)


def registryIncept(args):
    name = args.name
    alias = args.alias
    bran = args.bran
    base = args.base
    registryName = args.registry_name if args.registry_name is not None else alias
    nonce = args.nonce
    estOnly = args.establishment_only
    noBackers = args.no_backers
    backers = args.backers
    usage = args.usage

    if noBackers and backers:
        print("--no-backers and --backers can not both be provided")
        return -1

    icpDoer = RegistryInceptor(name=name, base=base, alias=alias, bran=bran, registryName=registryName,
                               nonce=nonce, estOnly=estOnly, noBackers=noBackers, baks=backers, usage=usage)

    doers = [icpDoer]
    return doers


class RegistryInceptor(doing.DoDoer):
    """ Creates a registry

    """

    def __init__(self, name, base, alias, bran, registryName, usage, **kwa):
        """  Create RegistryIncepter to pass message and process cues

        Parameters:
            name (str): name of habery and shared db and file path
            base (str): optional if "" path component of shared db and files.
            alias(str): human readable name of identifier to use for registry inception
            bran (str): Base64 22 char string that is used as base material for
                        seed. bran allows alphanumeric passcodes generated by key managers
                        like 1password to be key store for seed.
            registryName(str): human readable name for newly created registry

        """
        self.name = name
        self.alias = alias
        self.registryName = registryName
        self.usage = usage
        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.rgy = credentialing.Regery(hby=self.hby, name=name, base=base)
        self.hbyDoer = habbing.HaberyDoer(habery=self.hby)  # setup doer
        counselor = grouping.Counselor(hby=self.hby)
        self.postman = forwarding.Poster(hby=self.hby)

        notifier = Notifier(self.hby)
        mux = grouping.Multiplexor(self.hby, notifier=notifier)
        exc = exchanging.Exchanger(hby=self.hby, handlers=[])
        grouping.loadHandlers(exc, mux)

        mbx = indirecting.MailboxDirector(hby=self.hby, topics=["/receipt", "/multisig", "/replay"], exc=exc)
        self.registrar = credentialing.Registrar(hby=self.hby, rgy=self.rgy, counselor=counselor)
        doers = [self.hbyDoer, counselor, self.registrar, self.postman, mbx]
        self.toRemove = list(doers)

        doers.extend([doing.doify(self.inceptDo, **kwa)])
        super(RegistryInceptor, self).__init__(doers=doers)

    def inceptDo(self, tymth, tock=0.0, **kwa):
        """ Process incoming messages to incept a credential registry

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        Returns:  doifiable Doist compatible generator method
        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        hab = self.hby.habByName(self.alias)
        if hab is None:
            raise ValueError(f"{self.alias} is not a valid AID alias")

        estOnly = "estOnly" in kwa and kwa["estOnly"]
        registry = self.rgy.makeRegistry(name=self.registryName, prefix=hab.pre, **kwa)

        rseal = SealEvent(registry.regk, "0", registry.regd)
        rseal = dict(i=rseal.i, s=rseal.s, d=rseal.d)
        if estOnly:
            anc = hab.rotate(data=[rseal])
        else:
            anc = hab.interact(data=[rseal])

        aserder = serdering.SerderKERI(raw=bytes(anc))  # coring.Serder(raw=bytes(anc))
        self.registrar.incept(iserder=registry.vcp, anc=aserder)

        if isinstance(hab, GroupHab):
            usage = self.usage
            if usage is None:
                usage = input(f"Please enter a description of the credential registry: ")

            smids = hab.db.signingMembers(pre=hab.pre)
            smids.remove(hab.mhab.pre)

            for recp in smids:  # this goes to other participants only as a signaling mechanism
                exn, atc = grouping.multisigRegistryInceptExn(ghab=hab, vcp=registry.vcp.raw, anc=anc, usage=usage)
                self.postman.send(src=hab.mhab.pre,
                                  dest=recp,
                                  topic="multisig",
                                  serder=exn,
                                  attachment=atc)

        while not self.registrar.complete(pre=registry.regk, sn=0):
            self.rgy.processEscrows()
            yield self.tock

        print("Registry:  {}({}) \n\tcreated for Identifier Prefix:  {}".format(self.registryName,
                                                                                registry.regk, hab.pre))

        self.remove(self.toRemove)



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse
from ordered_set import OrderedSet as oset

from keri import help
from hio.base import doing

from keri.app import habbing, forwarding, grouping
from keri.app.cli.common import existing
from keri.core.coring import Ilks
from keri.core import serdering

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Notify other participants of the last event in a group multisig AID')
parser.set_defaults(handler=lambda args: handler(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the new identifier prefix', required=True)
parser.add_argument("--config", "-c", help="directory override for configuration data")

# Authentication for keystore
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument('--aeid', help='qualified base64 of non-transferable identifier prefix for  authentication '
                                   'and encryption of secrets in keystore', default=None)


def handler(args):
    """
    Send the /multisig/rot EXN notification message to other members of a multisig group about the last event in the KEL
    Args:
        args(Namespace): arguments object from command line
    """

    name = args.name
    base = args.base
    bran = args.bran
    alias = args.alias

    noticeDoer = NoticeDoer(name=name, base=base, alias=alias, bran=bran)

    doers = [noticeDoer]
    return doers


class NoticeDoer(doing.DoDoer):
    """ DoDoer Sending EXN notices to other participants.
    """

    def __init__(self, name, base, alias, bran):

        hby = existing.setupHby(name=name, base=base, bran=bran)
        self.hbyDoer = habbing.HaberyDoer(habery=hby)  # setup doer
        self.alias = alias
        self.hby = hby

        self.postman = forwarding.Postman(hby=self.hby)
        doers = [self.hbyDoer, self.postman, doing.doify(self.noticeDo)]

        super(NoticeDoer, self).__init__(doers=doers)

    def noticeDo(self, tymth, tock=0.0):
        """
        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        Returns:  doifiable Doist compatible generator method
        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        hab = self.hby.habByName(name=self.alias)

        if hab.group:
            (smids, rmids) = hab.members()
            serder = hab.kever.serder
            rot = hab.makeOwnEvent(sn=hab.kever.sn)
            eserder = serdering.SerderKERI(raw=rot)
            del rot[:eserder.size]

            ilk = serder.ked['t']
            others = list(oset(smids + (rmids or [])))  # list(rec.smids)
            others.remove(hab.mhab.pre)

            if ilk in (Ilks.rot,):
                print(f"Sending rot event to {len(others)} participants.")
                exn, ims = grouping.multisigRotateExn(hab,
                                                      aids=smids,
                                                      smids=smids,
                                                      rmids=rmids,
                                                      ked=serder.ked)
            elif ilk in (Ilks.icp,):
                print(f"Sending icp event to {len(others)} participants.")
                exn, ims = grouping.multisigInceptExn(hab,
                                                      aids=smids,
                                                      ked=serder.ked)
            elif ilk in (Ilks.ixn,):
                print(f"Sending ixn event to {len(others)} participants.")
                exn, ims = grouping.multisigInteractExn(hab,
                                                        aids=smids,
                                                        sn=serder.sn,
                                                        data=serder.ked["a"])
            else:
                raise ValueError(f"unsupport event type={ilk}")

            for recpt in others:
                self.postman.send(src=hab.mhab.pre, dest=recpt, topic="multisig",
                                  serder=eserder, attachment=rot)
                self.postman.send(src=hab.mhab.pre,
                                  dest=recpt,
                                  topic="multisig",
                                  serder=exn,
                                  attachment=ims)

            while len(self.postman.cues) != (2 * len(others)):
                yield self.tock

            print(f"Notice for event #{serder.sn} ({serder.ked['t']}) sent.")
            toRemove = [self.hbyDoer, self.postman]
            self.remove(toRemove)

        return



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse
import json

from keri import help
from hio.base import doing

from keri import kering
from keri.app.cli.common import existing

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Add new endpoint role authorization.')
parser.set_defaults(handler=lambda args: add_end(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the new identifier prefix', required=True)
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument("--aid", help="qualified base64 of AID to export rpy messages for all endpoints.",
                    required=True)


def add_end(args):
    """ Command line tool for adding endpoint role authorizations

    """
    ld = RoleDoer(name=args.name,
                  base=args.base,
                  alias=args.alias,
                  bran=args.bran,
                  aid=args.aid)
    return [ld]


class RoleDoer(doing.DoDoer):

    def __init__(self, name, base, alias, bran, aid):
        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.hab = self.hby.habByName(alias)
        if self.hab is None:
            raise kering.ConfigurationError(f"unknown alias={alias}")

        self.aid = aid
        doers = [doing.doify(self.roleDo)]

        super(RoleDoer, self).__init__(doers=doers)

    def roleDo(self, tymth, tock=0.0):
        """ Export any end reply messages previous saved for the provided AID

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        Returns:  doifiable Doist compatible generator method

        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        ends = self.hab.endsFor(self.aid)
        print(json.dumps(ends, indent=1))
        return



# -*- encoding: utf-8 -*-
"""
keri.kli.commands.multisig module

"""

import argparse
from ordered_set import OrderedSet as oset

from keri import help
from hio.base import doing

from keri import kering
from keri.app import grouping, indirecting, habbing, forwarding
from keri.app.cli.common import existing, displaying, config
from keri.app.notifying import Notifier
from keri.core import coring, serdering
from keri.peer import exchanging

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Begin or join a rotation of a group identifier')
parser.set_defaults(handler=lambda args: interactGroupIdentifier(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for the local identifier prefix', required=True)
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument('--data', '-d', help='Anchor data, \'@\' allowed', default=[], action="store", required=True)
parser.add_argument("--aids", "-g", help="List of other participant qb64 identifiers to include in interaction event",
                    action="append", required=False, default=None)


def interactGroupIdentifier(args):
    """
    Performs an interaction event on the group identifier specified as an argument.  The identifier prefix of the
    environment represented by the name parameter must be a member of the group identifier.  This command will
    perform an interaction of the local identifier if the sequence number of the local identifier is the same as the
    group identifier sequence number.  It will wait for all other members of the group to achieve the same sequence
    number (group + 1) and then publish the signed interaction event for the group identifier to all witnesses and
    wait for receipts.

    Parameters:
        args (parseargs):  command line parameters

    """

    data = config.parseData(args.data) if args.data is not None else None
    ixnDoer = GroupMultisigInteract(name=args.name, alias=args.alias, aids=args.aids, base=args.base, bran=args.bran,
                                    data=data)

    doers = [ixnDoer]
    return doers


class GroupMultisigInteract(doing.DoDoer):
    """
    Command line DoDoer to launch the needed coroutines to run launch Multisig interaction.
       This DoDoer will remove the multisig coroutine and exit when it receives a message
       that the multisig coroutine has successfully completed a cooperative rotation.

       ToDo: NRR
       Add .rmids and .smids

    """

    def __init__(self, name, alias, aids, base, bran, data):
        self.base = base
        self.bran = bran
        self.alias = alias
        self.aids = aids
        self.data = data

        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.hbyDoer = habbing.HaberyDoer(habery=self.hby)  # setup doer
        self.postman = forwarding.Poster(hby=self.hby)

        notifier = Notifier(self.hby)
        mux = grouping.Multiplexor(self.hby, notifier=notifier)
        exc = exchanging.Exchanger(hby=self.hby, handlers=[])
        grouping.loadHandlers(exc, mux)

        mbd = indirecting.MailboxDirector(hby=self.hby, topics=['/receipt', '/multisig'], exc=exc)
        self.counselor = grouping.Counselor(hby=self.hby)

        doers = [self.hbyDoer, self.postman, mbd, self.counselor]
        self.toRemove = list(doers)

        doers.extend([doing.doify(self.interactDo)])

        super(GroupMultisigInteract, self).__init__(doers=doers)

    def interactDo(self, tymth, tock=0.0):
        """ Create or participate in an interaction event for a distributed multisig identifier

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        ghab = self.hby.habByName(name=self.alias)
        if ghab is None:
            raise kering.ConfigurationError(f"invalid alias {self.alias} specified for database {self.hby.name}")

        aids = self.aids if self.aids is not None else ghab.smids

        ixn = ghab.interact(data=self.data)
        serder = serdering.SerderKERI(raw=ixn)

        exn, ims = grouping.multisigInteractExn(ghab=ghab, aids=aids, ixn=ixn)
        others = list(oset(ghab.smids + (ghab.rmids or [])))
        others.remove(ghab.mhab.pre)

        for recpt in others:  # send notification to other participants as a signalling mechanism
            self.postman.send(src=ghab.mhab.pre, dest=recpt, topic="multisig", serder=exn, attachment=ims)

        prefixer = coring.Prefixer(qb64=ghab.pre)
        seqner = coring.Seqner(sn=serder.sn)
        saider = coring.Saider(qb64b=serder.saidb)
        self.counselor.start(prefixer=prefixer, seqner=seqner, saider=saider, ghab=ghab)

        while True:
            saider = self.hby.db.cgms.get(keys=(prefixer.qb64, seqner.qb64))
            if saider is not None:
                break

            yield self.tock

        print()
        displaying.printIdentifier(self.hby, ghab.pre)
        self.remove(self.toRemove)



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.witness module

Witness command line interface
"""

import argparse
import logging
import os

from hio.base import doing

from keri.app import habbing, indirecting, configing
from keri.core.coring import Salter
from keri import help

parser = argparse.ArgumentParser(description="Run a demo collection of witnesses")
parser.add_argument("--loglevel", action="store", required=False, default=os.getenv("KERI_LOG_LEVEL", "CRITICAL"),
                    help="Set log level to DEBUG | INFO | WARNING | ERROR | CRITICAL. Default is CRITICAL")
parser.set_defaults(handler=lambda args: demo(args))


logger = help.ogler.getLogger()


def demo(args):
    """
    Run set of three witnesses for demo

    """
    base_formatter = logging.Formatter(
        '%(asctime)s [keri] %(module)s.%(funcName)s-%(lineno)s %(levelname)-8s %(message)s')
    base_formatter.default_msec_format = None
    help.ogler.baseConsoleHandler.setFormatter(base_formatter)
    help.ogler.level = logging.getLevelName(args.loglevel.upper())
    logger.setLevel(help.ogler.level)
    help.ogler.reopen(name="keri", temp=True, clear=True)

    wancf = configing.Configer(name="wan", headDirPath="scripts", temp=False, reopen=True, clear=False)
    wilcf = configing.Configer(name="wil", headDirPath="scripts", temp=False, reopen=True, clear=False)
    wescf = configing.Configer(name="wes", headDirPath="scripts", temp=False, reopen=True, clear=False)
    witcf = configing.Configer(name="wit", headDirPath="scripts", temp=False, reopen=True, clear=False)
    wubcf = configing.Configer(name="wub", headDirPath="scripts", temp=False, reopen=True, clear=False)
    wyzcf = configing.Configer(name="wyz", headDirPath="scripts", temp=False, reopen=True, clear=False)

    wanHby = habbing.Habery(name="wan", salt=Salter(raw=b'wann-the-witness').qb64, temp=False, cf=wancf)
    wilHby = habbing.Habery(name="wil", salt=Salter(raw=b'will-the-witness').qb64, temp=False, cf=wilcf)
    wesHby = habbing.Habery(name="wes", salt=Salter(raw=b'wess-the-witness').qb64, temp=False, cf=wescf)
    witHby = habbing.Habery(name="wit", salt=Salter(raw=b'witn-the-witness').qb64, temp=False, cf=witcf)
    wubHby = habbing.Habery(name="wub", salt=Salter(raw=b'wubl-the-witness').qb64, temp=False, cf=wubcf)
    wyzHby = habbing.Habery(name="wyz", salt=Salter(raw=b'wyzs-the-witness').qb64, temp=False, cf=wyzcf)

    doers = [InitDoer(wan=wanHby, wil=wilHby, wes=wesHby, wit=witHby, wub=wubHby, wyz=wyzHby)]

    return doers


class InitDoer(doing.DoDoer):

    def __init__(self, wan, wil, wes, wit, wub, wyz):
        self.wan = wan
        self.wil = wil
        self.wes = wes
        self.wit = wit
        self.wub = wub
        self.wyz = wyz
        
        super(InitDoer, self).__init__(doers=[doing.doify(self.initialize)])

    def initialize(self, tymth, tock=0.0):
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        wanDoers = indirecting.setupWitness(alias="wan", hby=self.wan, tcpPort=5632, httpPort=5642)
        wilDoers = indirecting.setupWitness(alias="wil", hby=self.wil, tcpPort=5633, httpPort=5643)
        wesDoers = indirecting.setupWitness(alias="wes", hby=self.wes, tcpPort=5634, httpPort=5644)

        witDoers = indirecting.setupWitness(alias="wit", hby=self.wit, tcpPort=5635, httpPort=5645)
        wubDoers = indirecting.setupWitness(alias="wub", hby=self.wub, tcpPort=5636, httpPort=5646)
        wyzDoers = indirecting.setupWitness(alias="wyz", hby=self.wyz, tcpPort=5637, httpPort=5647)

        self.extend(wanDoers + wilDoers + wesDoers + witDoers + wubDoers + wyzDoers)



# -*- encoding: utf-8 -*-
"""
keri.core.serdering module

"""
import copy
import json
from collections import namedtuple

import cbor2 as cbor
import msgpack
import pysodium
import blake3
import hashlib

from .. import kering
from ..kering import (ValidationError,  MissingFieldError,
                      ShortageError, VersionError, ProtocolError, KindError,
                      DeserializeError, FieldError, SerializeError)
from ..kering import (Versionage, Version, Vrsn_1_0, Vrsn_1_1,
                      VERRAWSIZE, VERFMT, VERFULLSIZE)
from ..kering import Protos, Serials, Rever, versify, deversify, Ilks
from ..core import coring
from .coring import MtrDex, DigDex, PreDex, Saids,  Digestage
from .coring import Matter, Saider, Verfer, Diger, Number, Tholder

from .. import help
from ..help import helping

logger = help.ogler.getLogger()

"""
Fieldage
    saids (dict): keyed by saidive field labels with values as default codes
    alls (dict): keyed by all field labels including saidive ones
                   with values as default codes

Example:
    Fields = Labelage(saids={'d': DigDex.Blake3_256},
                      alls={'v': '','d':''})
"""
Fieldage = namedtuple("Fieldage", "saids alls")  #values are dicts


"""
Reapage
    proto (str): protocol type value of Protos examples 'KERI', 'ACDC'
    major (str): single char hex string of major version number
    minor (str): single char hex string of minor version number
    kind (str): serialization value of Serials examples 'JSON', 'CBOR', 'MGPK'

"""
Reapage = namedtuple("Reapage", "proto major minor kind size")


class Serdery:
    """Serder factory class for generating serder instances by protocol type
    from an incoming message stream.


    """

    def __init__(self, *, version=None):
        """Init instance

        Parameters:
            version (Versionage | None): instance supported protocol version
                     None means do not enforce a supported version
        """
        self.version = version  # default version


    def reap(self, ims, *, version=None):
        """Extract and return Serder subclass based on protocol type reaped from
        version string inside serialized raw of Serder.

        Returns:
            serder (Serder): instance of Serder subclass where subclass is
                determined by the protocol type of its version string.

        Parameters:
            ims (bytearray) of serialized incoming message stream. Assumes start
                of stream is raw Serder.
            version (Versionage | None): instance supported protocol version
                None means do not enforce a supported version
        """
        version = version if version is not None else self.version

        if len(ims) < Serder.InhaleSize:
            raise ShortageError(f"Need more raw bytes for Serdery to reap.")

        match = Rever.search(ims)  # Rever regex takes bytes/bytearray not str
        if not match or match.start() > Serder.MaxVSOffset:
            raise VersionError(f"Invalid version string for Serder raw = "
                               f"{ims[: Serder.InhaleSize]}.")

        reaped = Reapage(*match.group("proto", "major", "minor", "kind", "size"))

        if reaped.proto == Protos.keri.encode("utf-8"):
            return SerderKERI(raw=ims, strip=True, version=version, reaped=reaped)
        elif reaped.proto == Protos.acdc.encode("utf-8"):
            return SerderACDC(raw=ims, strip=True, version=version, reaped=reaped)
        else:
            raise ProtocolError(f"Unsupported protocol type = {reaped.proto}.")



class Serder:
    """Serder is serializer-deserializer class for saidified  over-the-wire
    messages that deserializes to a field map (label value pairs) from
    either a serialized field map or an unlabeled fixed field structure with
    affiliated label list. The messages must include a version string field
    with proto (protocol), version, kind, and size elements or equivalent
    header with same elements. The messages may have an optional ilk field
    that is protocol specific. Protocols that have fixed  top-level fields
    also perform label inclusion validation.

    Message saidification and verification may be dependent on protocol and
    optionally ilk specific field label(s) and digest code type for its SAID(s).

    The base Serder class provides the common properties for all messages for
    all protocols. Each subclass is protocol based and adds properties that are
    required for all message ilks in a given protocol. Each protocol subclass
    may have dynamically injected ilk specific properties (as descriptors)
    if any.

    To support a new protocol with its ilks, add a protocol specific subclass,
    override the Labels class attribute, and if necessary the .verify and
    .saidify methods and define any protocol specific properties. If necessary
    define ilk specific subclasses of a given protocol (ilk is packet type).
    The .Labels class attributes configures the field label(s) for said
    generation and verification in addition to the required fields.

    Class Attributes:
        MaxVSOffset (int): Maximum Version String Offset in bytes/chars
        InhaleSize (int): Minimum raw buffer size needed to inhale
        Labels (dict): Protocol specific dict of field labels keyed by ilk
            (packet type string value). None is default key when no ilk needed.
            Each entry is a

    Properties:
        raw (bytes): of serialized event only
        sad (dict): self addressed data dict
        proto (str): Protocolage value as protocol identifier such as KERI, ACDC
        version (Versionage): protocol version (Major, Minor)
        kind (str): serialization kind coring.Serials such as JSON, CBOR, MGPK, CESR
        size (int): number of bytes in serialization
        said (str): qb64 said of .raw given by appropriate field
        saidb (bytes): qb64b of .said
        ilk (str | None): packet type for this Serder if any (may be None)


    Hidden Attributes:
        ._raw is bytes of serialized event only
        ._sad is key event dict
        ._proto (str):  Protocolage value as protocol type identifier
        ._version is Versionage instance of event version
        ._kind is serialization kind string value (see namedtuple coring.Serials)
            supported kinds are 'json', 'cbor', 'msgpack', 'binary'
        ._size is int of number of bytes in serialed event only
        ._said (str): qb64 given by appropriate saidive field

    Methods:
        verify()
        _verify()
        makify()
        compare()
        pretty(size: int | None ) -> str: Prettified JSON of this SAD

    ClassMethods:
        _inhale()
        _exhale()

    StaticMethods:
        loads()
        dumps()

    Note:
        loads and jumps of json use str whereas cbor and msgpack use bytes

    """

    MaxVSOffset = 12
    InhaleSize = MaxVSOffset + VERFULLSIZE  # min buffer size to inhale

    Dummy = "#"  # dummy spaceholder char for said. Must not be a valid Base64 char

    # should be same set of codes as in coring.DigestCodex coring.DigDex so
    # .digestive property works. Use unit tests to ensure codex sets match
    Digests = {
        DigDex.Blake3_256: Digestage(klas=blake3.blake3, size=None, length=None),
        DigDex.Blake2b_256: Digestage(klas=hashlib.blake2b, size=32, length=None),
        DigDex.Blake2s_256: Digestage(klas=hashlib.blake2s, size=None, length=None),
        DigDex.SHA3_256: Digestage(klas=hashlib.sha3_256, size=None, length=None),
        DigDex.SHA2_256: Digestage(klas=hashlib.sha256, size=None, length=None),
        DigDex.Blake3_512: Digestage(klas=blake3.blake3, size=None, length=64),
        DigDex.Blake2b_512: Digestage(klas=hashlib.blake2b, size=None, length=None),
        DigDex.SHA3_512: Digestage(klas=hashlib.sha3_512, size=None, length=None),
        DigDex.SHA2_512: Digestage(klas=hashlib.sha512, size=None, length=None),
    }

    #override in subclass to enforce specific protocol
    Protocol = None  # required protocol, None means any in Protos is ok

    Proto = Protos.keri  # default protocol type
    Vrsn = Vrsn_1_0  # default protocol version for protocol type
    Kind = Serials.json  # default serialization kind


    # Nested dict keyed by protocol.
    # Each protocol value is a dict keyed by ilk.
    # Each ilk value is a Labelage named tuple with saids, codes and fields
    # ilk value of None is default for protocols that support ilkless packets
    Fields = {
            Protos.keri:
            {
                Vrsn_1_0:
                {
                    Ilks.icp: Fieldage(saids={Saids.d: DigDex.Blake3_256,
                                              Saids.i: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', s='0', kt='0',
                            k=[], nt='0', n=[], bt='0', b=[], c=[], a=[])),
                    Ilks.rot: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', i='', s='0', p='',
                            kt='0',k=[], nt='0', n=[], bt='0', br=[],
                            ba=[], a=[])),
                    Ilks.ixn: Fieldage({Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', i='', s='0', p='', a=[])),
                    Ilks.dip: Fieldage(saids={Saids.d: DigDex.Blake3_256,
                                              Saids.i: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', s='0', kt='0',
                            k=[], nt='0', n=[], bt='0', b=[], c=[], a=[],
                            di='')),
                    Ilks.drt: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', i='', s='0', p='',
                            kt='0',k=[], nt='0', n=[], bt='0', br=[],
                            ba=[], a=[])),
                    Ilks.rct: Fieldage(saids={},
                        alls=dict(v='', t='',d='', i='', s='0')),
                    Ilks.qry: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', dt='', r='', rr='',
                                    q={})),
                    Ilks.rpy: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', dt='', r='',a=[])),
                    Ilks.pro: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', dt='', r='', rr='',
                                    q={})),
                    Ilks.bar: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', dt='', r='',a=[])),
                    Ilks.exn: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='', d='', i="", rp='', p="", dt='', r='',q={},
                                    a=[], e={})),
                    Ilks.vcp: Fieldage(saids={Saids.d: DigDex.Blake3_256,
                                              Saids.i: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', ii='', s='0', c=[],
                                    bt='0', b=[], n='')),
                    Ilks.vrt: Fieldage(saids={Saids.d: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', p='', s='0',
                                    bt='0', br=[], ba=[])),
                    Ilks.iss: Fieldage(saids={Saids.d: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', s='0', ri='',
                                  dt='')),
                    Ilks.rev: Fieldage(saids={Saids.d: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', s='0', ri='',
                                  p='', dt='')),
                    Ilks.bis: Fieldage(saids={Saids.d: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', ii='', s='0', ra={},
                                  dt='')),
                    Ilks.brv: Fieldage(saids={Saids.d: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', s='0', p='', ra={},
                                  dt='')),
                },
                Vrsn_1_1:
                {
                    Ilks.icp: Fieldage(saids={Saids.d: DigDex.Blake3_256,
                                              Saids.i: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', s='0', kt='0',
                            k=[], nt='0', n=[], bt='0', b=[], c=[], a=[])),
                    Ilks.rot: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', i='', s='0', p='',
                            kt='0',k=[], nt='0', n=[], bt='0', br=[],
                            ba=[], c=[], a=[])),
                    Ilks.ixn: Fieldage({Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', i='', s='0', p='', a=[])),
                    Ilks.dip: Fieldage(saids={Saids.d: DigDex.Blake3_256,
                                              Saids.i: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', s='0', kt='0',
                            k=[], nt='0', n=[], bt='0', b=[], c=[], a=[],
                            di='')),
                    Ilks.drt: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', i='', s='0', p='',
                            kt='0',k=[], nt='0', n=[], bt='0', br=[],
                            ba=[], c=[], a=[])),
                    Ilks.rct: Fieldage(saids={},
                        alls=dict(v='', t='',d='', i='', s='0')),
                    Ilks.qry: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', i='', dt='', r='', rr='',
                                    q={})),
                    Ilks.rpy: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', i='', dt='', r='',a=[])),
                    Ilks.pro: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', i='', dt='', r='', rr='',
                                    q={})),
                    Ilks.bar: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='',d='', i='', dt='', r='',a=[])),
                    Ilks.exn: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                        alls=dict(v='', t='', d='', i="", rp="", p="", dt='', r='', q={},
                                    a=[], e={})),
                },
            },
            Protos.crel:
            {
                Vrsn_1_1:
                {
                    Ilks.vcp: Fieldage(saids={Saids.d: DigDex.Blake3_256,
                                              Saids.i: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', ii='', s='0', c=[],
                                    bt='0', b=[], u='')),
                    Ilks.vrt: Fieldage(saids={Saids.d: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', p='', s='0',
                                    bt='0', br=[], ba=[])),
                    Ilks.iss: Fieldage(saids={Saids.d: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', s='0', ri='',
                                  dt='')),
                    Ilks.rev: Fieldage(saids={Saids.d: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', s='0', ri='',
                                  p='', dt='')),
                    Ilks.bis: Fieldage(saids={Saids.d: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', ii='', s='0', ra={},
                                  dt='')),
                    Ilks.brv: Fieldage(saids={Saids.d: DigDex.Blake3_256,},
                        alls=dict(v='', t='',d='', i='', s='0', p='', ra={},
                                  dt='')),
                },
            },
            Protos.acdc:
            {
                Vrsn_1_0:
                {
                    None: Fieldage(saids={Saids.d: DigDex.Blake3_256},
                                   alls=dict(v='', d='', i='', s='')),
                }
            },
        }


    # default ilk for each protocol at default version is zeroth ilk in dict
    Ilks = dict()
    for key, val in Fields.items():
        Ilks[key] = list(list(val.values())[0].keys())[0]


    def __init__(self, *, raw=b'', sad=None, strip=False, version=Version,
                 reaped=None, verify=True, makify=False,
                 proto=None, vrsn=None, kind=None, ilk=None, saids=None):
        """Deserialize raw if provided. Update properties from deserialized raw.
            Verifies said(s) embedded in sad as given by labels.
            When verify is True then verify said(s) in deserialized raw as
            given by label(s) according to proto and ilk and code
        If raw not provided then serialize .raw from sad with kind and code.
            When kind not provided use kind embedded in sad['v'] version string.
            When saidify is True then compute and update said(s) in sad as
            given by label(s) according to proto and ilk and code.

        Parameters:
            raw (bytes): serialized event
            sad (dict): serializable saidified field map of message.
                Ignored if raw provided
            strip (bool): True means strip (delete) raw from input stream
                bytearray after parsing. False means do not strip.
                Assumes that raw is bytearray when strip is True.
            version (Versionage | None): instance supported protocol version
                None means do not enforce a supported version
            reaped (Reapage | None): instance of deconstructed version string
                elements. If none or empty ignore otherwise assume that raw
                already had its version string extracted (reaped) into the
                elements of reaped.
            verify (bool): True means verify said(s) of given raw or sad.
                Raises ValidationError if verification fails
                Ignore when raw not provided or when raw and saidify is True
            makify (bool): True means compute fields for sad including size and
                saids.
            proto (str | None): desired protocol type str value of Protos
                If None then its extracted from sad or uses default .Proto
            vrsn (Versionage | None): instance desired protocol version
                If None then its extracted from sad or uses default .Vrsn
            kind (str None): serialization kind string value of Serials
                supported kinds are 'json', 'cbor', 'msgpack', 'binary'
                If None then its extracted from sad or uses default .Kind
            ilk (str | None): desired ilk packet type str value of Ilks
                If None then its extracted from sad or uses default .Ilk
            saids (dict): of keyed by label of codes for saidive fields to
                override defaults given in .Fields for a given ilk.
                If None then use defaults
                Code assignment for each saidive field in desending priority:
                   - the code provided in saids when not None
                   - the code extracted from sad[said label] when valid CESR
                   - the code provided in .Fields...saids


        """

        if raw:  # deserialize raw using property setter
            # self._inhale works because it only references class attributes
            sad, proto, vrsn, kind, size = self._inhale(raw=raw,
                                                        version=version,
                                                        reaped=reaped)
            self._raw = bytes(raw[:size])  # crypto ops require bytes not bytearray
            self._sad = sad
            self._proto = proto
            self._vrsn = vrsn
            self._kind = kind
            self._size = size
            # primary said field label
            try:
                label = list(self.Fields[self.proto][self.vrsn][self.ilk].saids.keys())[0]
                if label not in self._sad:
                    raise FieldError(f"Missing primary said field in {self._sad}.")
                self._said = self._sad[label]  # not verified
            except Exception as ex:
                self._said = None  # no saidive field

            if strip:  #only when raw is bytearray
                try:
                    del raw[:self._size]
                except TypeError:
                    pass  # ignore if bytes

            if verify:  # verify the said(s) provided in raw
                try:
                    self._verify()  # raises exception when not verify
                except Exception as ex:
                    logger.error("Invalid raw for Serder %s\n%s",
                                 self.pretty(), ex.args[0])
                    raise ValidationError(f"Invalid raw for Serder = "
                                          f"{self._sad}. {ex.args[0]}") from ex

        elif sad or makify:  # serialize sad into raw or make sad
            if makify:  # recompute properties and said(s) and reset sad
                # makify resets sad, raw, proto, version, kind, and size
                self.makify(sad=sad, version=version,
                        proto=proto, vrsn=vrsn, kind=kind, ilk=ilk, saids=saids)

            else:
                # self._exhale works because it only access class attributes
                raw, sad, proto, vrsn, kind, size = self._exhale(sad=sad,
                                                                 version=version)
                self._raw = raw
                self._sad = sad
                self._proto = proto
                self._vrsn = vrsn
                self._kind = kind
                self._size = size
                # primary said field label
                try:
                    label = list(self.Fields[self.proto][self.vrsn][self.ilk].saids.keys())[0]
                    if label not in self._sad:
                        raise DeserializeError(f"Missing primary said field in {self._sad}.")
                    self._said = self._sad[label]  # not verified
                except Exception:
                    self._said = None  # no saidive field

                if verify:  # verify the said(s) provided in sad
                    try:
                        self._verify()  # raises exception when not verify
                    except Exception as ex:
                        logger.error("Invalid sad for Serder %s\n%s",
                                     self.pretty(), ex.args[0])
                        raise ValidationError(f"Invalid sad for Serder ="
                                              f"{self._sad}.") from ex

        else:
            raise ValueError("Improper initialization need raw or sad or makify.")



    def verify(self):
        """Verifies said(s) in sad against raw
        Override for protocol and ilk specific verification behavior. Especially
        for inceptive ilks that have more than one said field like a said derived
        identifier prefix.

        Returns:
            verify (bool): True if said(s) verify. False otherwise
        """
        try:
            self._verify()
        except Exception as ex:              # log validation error here
            logger.error("Invalid Serder: %s\n for %s\n",
                         ex.args[0], self.pretty())

            return False

        return True


    def _verify(self):
        """Verifies said(s) in sad against raw
        Override for protocol and ilk specific verification behavior. Especially
        for inceptive ilks that have more than one said field like a said derived
        identifier prefix.

        Raises a ValidationError (or subclass) if any verification fails

        """
        if self.Protocol and self.proto != self.Protocol:
            raise ValidationError(f"Expected protocol = {self.Protocol}, got "
                                 f"{self.proto} instead.")

        if self.proto not in self.Fields:
            raise ValidationError(f"Invalid protocol type = {self.proto}.")

        if self.ilk not in self.Fields[self.proto][self.vrsn]:
            raise ValidationError(f"Invalid packet type (ilk) = {self.ilk} for"
                                  f"protocol = {self.proto}.")

        fields = self.Fields[self.proto][self.vrsn][self.ilk]  # get labelage
        # ensure all required fields in alls are in sad
        alls = fields.alls  # dict of all field labels with default values
        keys = list(self._sad)  # get list of keys of self.sad
        for key in list(keys):  # make copy to mutate
            if key not in alls:
                del keys[keys.index(key)]  # remove non required fields

        if list(alls.keys()) != keys:  # forces ordering of labels in .sad
            raise MissingFieldError(f"Missing one or more required fields from"
                                    f"= {list(alls.keys())} in sad = "
                                    f"{self._sad}.")

        # said field labels are not order dependent with respect to all fields
        # in sad so use set() to test inclusion
        saids = copy.copy(fields.saids)  # get copy of saidive field labels and defaults values
        if not (set(saids.keys()) <= set(alls.keys())):
            raise MissingFieldError(f"Missing one or more required said fields"
                                    f" from {list(saids.keys())} in sad = "
                                    f"{self._sad}.")

        sad = self.sad  # make shallow copy so don't clobber original .sad
        for label in saids.keys():
            try:  # replace default code with code of value from sad
                saids[label] = Matter(qb64=sad[label]).code
            except Exception as ex:
                if saids[label] in DigDex:  # digestive but invalid
                    raise ValidationError(f"Invalid said field '{label}' in sad\n"
                                      f" = {self._sad}.") from ex

            if saids[label] in DigDex:  # if digestive then replace with dummy
                sad[label] = self.Dummy * len(sad[label])


        raw = self.dumps(sad, kind=self.kind)  # serialize dummied sad copy

        for label, code in saids.items():
            if code in DigDex:  # subclass override if non digestive allowed
                klas, size, length = self.Digests[code]  # digest algo size & length
                ikwa = dict()  # digest algo class initi keyword args
                if size:
                    ikwa.update(digest_size=size)  # optional digest_size
                dkwa = dict()  # digest method keyword args
                if length:
                    dkwa.update(length=length)
                dig = Matter(raw=klas(raw, **ikwa).digest(**dkwa), code=code).qb64
                if dig != self._sad[label]:  # compare to original
                    raise ValidationError(f"Invalid said field '{label}' in sad"
                                          f" = {self._sad}, should be {dig}.")
                sad[label] = dig

        raw = self.dumps(sad, kind=self.kind)
        if raw != self.raw:
            raise ValidationError(f"Invalid round trip of {sad} != \n"
                                  f"{self.sad}.")
        # verified successfully since no exception


    def makify(self, sad, *, version=None,
               proto=None, vrsn=None, kind=None, ilk=None, saids=None):
        """Makify given sad dict makes the versions string and computes the said
        field values and sets associated properties:
        raw, sad, proto, version, kind, size

        Override for protocol and ilk specific saidification behavior. Especially
        for inceptive ilks that have more than one said field like a said derived
        identifier prefix.

        Default prioritization.
           Use method parameter if not None
           Else use provided version string if valid
           Otherwise use class attribute


        Parameters:
            sad (dict): serializable saidified field map of message.
                Ignored if raw provided
            version (Versionage): instance supported protocol version
                None means do not enforce version
            proto (str | None): desired protocol type str value of Protos
                If None then its extracted from sad or uses default .Proto
            vrsn (Versionage | None): instance desired protocol version
                If None then its extracted from sad or uses default .Vrsn
            kind (str None): serialization kind string value of Serials
                supported kinds are 'json', 'cbor', 'msgpack', 'binary'
                If None then its extracted from sad or uses default .Kind
            ilk (str | None): desired ilk packet type str value of Ilks
                If None then its extracted from sad or uses default .Ilk
            saids (dict): of keyed by label of codes for saidive fields to
                override defaults given in .Fields for a given ilk.
                If None then use defaults
                Code assignment for each saidive field in desending priority:
                   - the code provided in saids when not None
                   - the code extracted from sad[said label] when valid CESR
                   - the code provided in .Fields...saids
        """
        sproto = svrsn = skind = silk = None
        if sad and 'v' in sad:  # attempt to get from vs in sad
            try:  # extract version string elements as defaults if provided
                sproto, svrsn, skind, _ = deversify(sad["v"], version=version)
            except ValueError as ex:
                pass
            else:
                silk = sad.get('t')  # if not in get returns None which may be valid

        if proto is None:
            proto = sproto if sproto is not None else self.Proto

        if vrsn is None:
            vrsn = svrsn if svrsn is not None else self.Vrsn

        if kind is None:
            kind = skind if skind is not None else self.Kind

        if ilk is None:
            ilk = silk if silk is not None else self.Ilks[proto]


        if proto not in self.Fields:
            raise SerializeError(f"Invalid protocol type = {proto}.")


        if self.Protocol and proto != self.Protocol:
            raise SerializeError(f"Expected protocol = {self.Protocol}, got "
                                 f"{proto} instead.")

        if version is not None and vrsn != version:
            raise SerializeError(f"Expected version = {version}, got "
                               f"{vrsn.major}.{vrsn.minor}.")

        if kind not in Serials:
            raise SerializeError(f"Invalid serialization kind = {kind}")


        if ilk not in self.Fields[proto][vrsn]:
            raise SerializeError(f"Invalid packet type (ilk) = {ilk} for"
                                  f"protocol = {proto}.")

        fields = self.Fields[proto][vrsn][ilk]  # get Fieldage of fields

        if not sad:  # empty or None so create from defaults
            sad = {}
            for label, value in fields.alls.items():
                if helping.nonStringIterable(value):  # copy iterable defaults
                    value = copy.copy(value)
                sad[label] = value

            if 't' in sad:  # packet type (ilk) requried so set value to ilk
                sad['t'] = ilk

        # ensure all required fields in alls are in sad
        alls = fields.alls  # all field labels
        for label, value in alls.items():  # ensure provided sad as all required fields
            if label not in sad:  # supply default
                if helping.nonStringIterable(value):  # copy iterable defaults
                    value = copy.copy(value)
                sad[label] = value

        keys = list(sad)  # get list of keys of self.sad
        for key in list(keys):  # make copy to mutate
            if key not in alls:
                del keys[keys.index(key)]  # remove non required fields

        if list(alls.keys()) != keys:  # ensure ordering of fields matches alls
            raise SerializeError(f"Mismatch one or more of all required fields "
                                 f" = {list(alls.keys())} in sad = {sad}.")

        # said field labels are not order dependent with respect to all fields
        # in sad so use set() to test inclusion
        _saids = copy.copy(fields.saids)  # get copy of defaults
        if not (set(_saids.keys()) <= set(alls.keys())):
            raise SerializeError(f"Missing one or more required said fields "
                                 f"from {list(_saids.keys())} in sad = {sad}.")

        # override saidive defaults
        for label in _saids:
            if saids and label in saids:  # use parameter override
                _saids[label] = saids[label]
            else:
                try:  # use sad field override
                    _saids[label] = Matter(qb64=sad[label]).code
                except Exception:
                    pass  # no override

            if _saids[label] in DigDex:  # if digestive then fill with dummy
                sad[label] = self.Dummy * Matter.Sizes[_saids[label]].fs


        if 'v' not in sad:  # ensures that 'v' is always required by .Labels
            raise SerializeError(f"Missing requires version string field 'v'"
                                          f" in sad = {sad}.")

        sad['v'] = self.Dummy * VERFULLSIZE  # ensure size of vs

        raw = self.dumps(sad, kind)  # get size of fully dummied sad
        size = len(raw)

        # generate new version string with correct size
        vs = versify(proto=proto, version=vrsn, kind=kind, size=size)
        sad["v"] = vs  # update version string in sad

        # now have correctly sized version string in sad
        # now compute saidive digestive field values using sized dummied sad
        raw = self.dumps(sad, kind=kind)  # serialize sized dummied sad

        for label, code in _saids.items():
            if code in DigDex:  # subclass override if non digestive allowed
                klas, dsize, dlen = self.Digests[code]  # digest algo size & length
                ikwa = dict()  # digest algo class initi keyword args
                if dsize:
                    ikwa.update(digest_size=dsize)  # optional digest_size
                dkwa = dict()  # digest method keyword args
                if dlen:
                    dkwa.update(length=dlen)
                dig = Matter(raw=klas(raw, **ikwa).digest(**dkwa), code=code).qb64
                sad[label] = dig

        raw = self.dumps(sad, kind=kind)  # compute final raw

        self._raw = raw
        self._sad = sad
        self._proto = proto
        self._vrsn = vrsn
        self._kind = kind
        self._size = size
        # primary said field label
        try:
            label = list(self.Fields[self.proto][self.vrsn][self.ilk].saids.keys())[0]
            if label not in self._sad:
                raise SerializeError(f"Missing primary said field in {self._sad}.")
            self._said = self._sad[label]  # implicitly verified
        except Exception:
            self._said = None  # no saidive field



    @classmethod
    def _inhale(clas, raw, version=Version, reaped=None):
        """Deserializes raw.
        Parses serilized event ser of serialization kind and assigns to
        instance attributes and returns tuple of associated elements.

        As classmethod enables testing parsing raw serder values. This can be
        called on self as well because it only ever accesses clas attributes
        not instance attributes.

        Returns: tuple (sad, proto, vrsn, kind, size) where:
            sad (dict): serializable attribute dict of saidified data
            proto (str): value of Protos (Protocolage) protocol type
            vrsn (Versionage | None): tuple of (major, minor) version ints
                None means do not enforce version
            kind (str): value of Serials (Serialage) serialization kind

        Parameters:
            raw (bytes): serialized sad message
            version (Versionage): instance supported protocol version
            reaped (Reapage | None): instance of deconstructed version string
                elements. If none or empty ignore otherwise assume that raw
                already had its version string extracted (reaped) into the
                elements of reaped.

        Note:
            loads and jumps of json use str whereas cbor and msgpack use bytes
            Assumes only supports Version

        """
        if reaped:
            proto, major, minor, kind, size = reaped  # tuple unpack
        else:
            if len(raw) < clas.InhaleSize:
                raise ShortageError(f"Need more raw bytes for Serder to inhale.")

            match = Rever.search(raw)  # Rever regex takes bytes/bytearray not str
            if not match or match.start() > clas.MaxVSOffset:
                raise VersionError(f"Invalid version string in raw = "
                                   f"{raw[:clas.InhaleSize]}.")

            proto, major, minor, kind, size = match.group("proto",
                                                          "major",
                                                          "minor",
                                                          "kind",
                                                          "size")

        proto = proto.decode("utf-8")
        if proto not in Protos:
            raise ProtocolError(f"Invalid protocol type = {proto}.")

        vrsn = Versionage(major=int(major, 16), minor=int(minor, 16))
        if version is not None and vrsn != version:
            raise VersionError(f"Expected version = {version}, got "
                               f"{vrsn.major}.{vrsn.minor}.")

        kind = kind.decode("utf-8")
        if kind not in Serials:
            raise KindError(f"Invalid serialization kind = {kind}.")

        size = int(size, 16)
        if len(raw) < size:
            raise ShortageError(f"Need more bytes.")

        sad = clas.loads(raw=raw, size=size, kind=kind)

        if "v" not in sad:
            raise FieldError(f"Missing version string field in {sad}.")

        return sad, proto, vrsn, kind, size


    @staticmethod
    def loads(raw, size=None, kind=Serials.json):
        """Utility static method to handle deserialization by kind

        Returns:
           sad (dict | list): deserialized dict or list. Assumes attribute
                dict of saidified data.

        Parameters:
           raw (bytes |bytearray): raw serialization to deserialze as dict
           size (int): number of bytes to consume for the deserialization.
                       If None then consume all bytes in raw
           kind (str): value of Serials (Serialage) serialization kind
                       "JSON", "MGPK", "CBOR"
        """
        if kind == Serials.json:
            try:
                sad = json.loads(raw[:size].decode("utf-8"))
            except Exception as ex:
                raise DeserializeError(f"Error deserializing JSON: "
                    f"{raw[:size].decode('utf-8')}") from ex

        elif kind == Serials.mgpk:
            try:
                sad = msgpack.loads(raw[:size])
            except Exception as ex:
                raise DeserializeError(f"Error deserializing MGPK: "
                    f"{raw[:size].decode('utf-8')}") from ex

        elif kind == Serials.cbor:
            try:
                sad = cbor.loads(raw[:size])
            except Exception as ex:
                raise DeserializeError(f"Error deserializing CBOR: "
                    f"{raw[:size].decode('utf-8')}") from ex

        else:
            raise DeserializeError(f"Invalid deserialization kind: {kind}")

        return sad


    @classmethod
    def _exhale(clas, sad, version=None):
        """Serializes sad given kind and version and sets the serialized size
        in the version string.

        As classmethod enables bootstrap of valid sad dict that has correct size
        in version string. This obviates sizeify. This can be called on self as
        well because it only ever accesses clas attributes not instance attributes.

        Returns tuple of (raw, proto, kind, sad, vrsn) where:
            raw (str): serialized event as bytes of kind
            proto (str): protocol type as value of Protocolage
            kind (str): serialzation kind as value of Serialage
            sad (dict): modified serializable attribute dict of saidified data
            vrsn (Versionage): tuple value (major, minor)

        Parameters:
            sad (dict): serializable attribute dict of saidified data
            version (Versionage | None): supported protocol version for message
                None means do not enforce a supported version


        """
        if "v" not in sad:
            raise SerializeError(f"Missing version string field in {sad}.")

        # extract elements so can replace size element but keep others
        proto, vrsn, kind, size = deversify(sad["v"], version=version)

        raw = clas.dumps(sad, kind)
        size = len(raw)

        # generate new version string with correct size
        vs = versify(proto=proto, version=vrsn, kind=kind, size=size)

        # find location of old version string inside raw
        match = Rever.search(raw)  # Rever's regex takes bytes
        if not match or match.start() > 12:
            raise SerializeError(f"Invalid version string in raw = {raw}.")
        fore, back = match.span()  # start and end positions of version string

        # replace old version string in raw with new one
        raw = b'%b%b%b' % (raw[:fore], vs.encode("utf-8"), raw[back:])
        if size != len(raw):  # substitution messed up
            raise SerializeError(f"Malformed size of raw in version string == {vs}")
        sad["v"] = vs  # update sad

        return raw, sad, proto, vrsn, kind, size


    @staticmethod
    def dumps(sad, kind=Serials.json):
        """Utility static method to handle serialization by kind

        Returns:
           raw (bytes): serialization of sad dict using serialization kind

        Parameters:
           sad (dict | list)): serializable dict or list to serialize
           kind (str): value of Serials (Serialage) serialization kind
                "JSON", "MGPK", "CBOR"
        """
        if kind == Serials.json:
            raw = json.dumps(sad, separators=(",", ":"),
                             ensure_ascii=False).encode("utf-8")

        elif kind == Serials.mgpk:
            raw = msgpack.dumps(sad)

        elif kind == Serials.cbor:
            raw = cbor.dumps(sad)
        else:
            raise SerializeError(f"Invalid serialization kind = {kind}")

        return raw


    def compare(self, said=None):
        """Utility method to allow comparison of own .said digest of .raw
        with some other purported said of .raw

        Returns:
            success (bool): True if said matches self.saidb  via string
               equality. Converts said to bytes if unicode


        Parameters:
            said (bytes | str): qb64b or qb64 digest to compare with .said
        """
        if said is not None:
            if hasattr(said, "encode"):
                said = said.encode('utf-8')  # makes bytes

            return said == self.saidb  # str match bool

        else:
            raise ValidationError(f"Uncomparable saids.")



    def pretty(self, *, size=None):
        """Utility method to pretty print .sad as JSON str.
        Returns:
            pretty (str):  JSON of .sad with pretty formatting

        Pararmeters:
            size (int | None): size limit. None means not limit.
                Enables protection against syslog error when
                exceeding UDP MTU (max trans unit) for syslog applications.
                Guaranteed IPv4 MTU is 576, and IPv6 MTU is 1280.
                Most broadband routers have an UDP MTU set to 1454.
                Must include not just payload but UDP/IP header in
                MTU calculation. So must leave room for either UDP/IpV4 or
                the bigger UDP/IPv6 header.
                Except for old IoT hardware, modern implementations all
                support IPv6 so 1024 is usually a safe value for payload.
        """
        return json.dumps(self._sad, indent=1)[:size]


    @property
    def raw(self):
        """raw property getter
        Returns:
            raw (bytes):  serialized version
        """
        return self._raw


    @property
    def sad(self):
        """sad property getter
        Returns:
            sad (dict): serializable attribute dict (saidified data)
        """
        return dict(self._sad)  # return copy


    @property
    def kind(self):
        """kind property getter
        Returns:
            kind (str): value of Serials (Serialage)"""
        return self._kind


    @property
    def proto(self):
        """proto property getter
        protocol identifier type value of Protocolage such as 'KERI' or 'ACDC'

        Returns:
            proto (str): Protocolage value as protocol type
        """
        return self._proto


    @property
    def vrsn(self):
        """vrsn (version) property getter

        Returns:
            vrsn (Versionage): instance of protocol version for this Serder
        """
        return self._vrsn

    @property
    def version(self):
        """version property getter alias of .vrsn

        Returns:
            version (Versionage): instance of protocol version for this Serder
        """
        return self.vrsn


    @property
    def size(self):
        """size property getter
        Returns:
            size (int): number of bytes in .raw
        """
        return self._size


    @property
    def said(self):
        """said property getter
        Returns:
           said (str): qb64
        """
        if not self.Fields[self.proto][self.vrsn][self.ilk].saids.keys() and 'd' in self._sad:
            return self._sad['d']  # special case for non-saidive messages like rct
        return self._said


    @property
    def saidb(self):
        """saidb property getter
        Returns:
            saidb (bytes): qb64b of said  of .saider
        """
        return self.said.encode("utf-8") if self.said is not None else None


    @property
    def ilk(self):
        """ilk property getter
        Returns:
            ilk (str): pracket type given by sad['t'] if any
        """
        return self._sad.get('t')  # returns None if 't' not in sad



class SerderKERI(Serder):
    """SerderKERI is Serder subclass with Labels for KERI packet types (ilks) and
       properties for exposing field values of KERI messages

       See docs for Serder
    """
    #override in subclass to enforce specific protocol
    Protocol = Protos.keri  # required protocol, None means any in Protos is ok
    Proto = Protos.keri  # default protocol type



    def _verify(self, **kwa):
        """Verifies said(s) in sad against raw
        Override for protocol and ilk specific verification behavior. Especially
        for inceptive ilks that have more than one said field like a said derived
        identifier prefix.

        Raises a ValidationError (or subclass) if any verification fails

        """
        super(SerderKERI, self)._verify(**kwa)

        allkeys = list(self.Fields[self.proto][self.vrsn][self.ilk].alls.keys())
        keys = list(self.sad.keys())
        if allkeys != keys:
            raise ValidationError(f"Invalid top level field list. Expected "
                                  f"{allkeys} got {keys}.")

        if (self.vrsn.major < 2 and self.vrsn.minor < 1 and
            self.ilk in (Ilks.qry, Ilks.rpy, Ilks.pro, Ilks.bar, Ilks.exn)):
                pass
        else:  # verify pre
            try:
                code = Matter(qb64=self.pre).code
            except Exception as ex:
                raise ValidationError(f"Invalid identifier prefix = "
                                      f"{self.pre}.") from ex

            if self.ilk in (Ilks.dip, Ilks.drt):
                idex = DigDex  # delegatee must be digestive prefix
            else:
                idex = PreDex  # non delegatee may be non digest

            if code not in idex:
                raise ValidationError(f"Invalid identifier prefix code = {code}.")

            # non-transferable pre validations
            if code in [PreDex.Ed25519N, PreDex.ECDSA_256r1N, PreDex.ECDSA_256k1N]:
                if self.ndigs:
                    raise ValidationError(f"Non-transferable code = {code} with"
                                          f" non-empty nxt = {self.ndigs}.")

                if self.backs:
                    raise ValidationError("Non-transferable code = {code} with"
                                          f" non-empty backers = {self.backs}.")

                if self.seals:
                    raise ValidationError("Non-transferable code = {code} with"
                                          f" non-empty seals = {self.seals}.")

        if self.ilk in (Ilks.dip):  # validate delpre
            try:
                code = Matter(qb64=self.delpre).code
            except Exception as ex:
                raise ValidationError(f"Invalid delegator prefix = "
                                      f"{self.delpre}.") from ex

            if code not in PreDex:  # delegator must be valid prefix code
                raise ValidationError(f"Invalid delegator prefix code = {code}.")


    @property
    def estive(self):  # establishative
        """ Returns True if Serder represents an establishment event """
        return (self._sad["t"] in (Ilks.icp, Ilks.rot, Ilks.dip, Ilks.drt)
                     if "t" in self._sad else False)


    @property
    def ked(self):
        """
        Returns:
            ked (dict): key event dict property getter. Alias for .sad
        """
        return self.sad


    @property
    def pre(self):
        """
        Returns:
           pre (str): qb64  of .sad["i"] identifier prefix property getter
        """
        return self._sad.get("i")


    @property
    def preb(self):
        """
        Returns:
        preb (bytes): qb64b  of .pre identifier prefix property getter as bytes
        """
        return self.pre.encode("utf-8") if self.pre is not None else None


    @property
    def sner(self):
        """Number instance of sequence number, sner property getter

        Returns:
            (Number): of ._sad["s"] hex number str converted
        """
        # auto converts hex num str to int
        return Number(num=self._sad["s"]) if 's' in self._sad else None


    @property
    def sn(self):
        """Sequence number, sn property getter
        Returns:
            sn (int): of .sner.num from .sad["s"]
        """
        return self.sner.num if self.sner is not None else None


    @property
    def snh(self):
        """Sequence number hex str, snh property getter
        Returns:
            snh (hex str): of .sner.numh from .sad["s"]
        """
        return self.sner.numh if self.sner is not None else None


    @property
    def seals(self):
        """Seals property getter

        Returns:
            seals (list): from ._sad["a"]
        """
        return self._sad.get("a")

    #Properties of inceptive Serders ilks in (icp, dip) and version2 estive serders

    @property
    def traits(self):
        """Traits list property getter  (config traits)

        Returns:
            traits (list): from ._sad["c"]
        """
        return self._sad.get("c")


    #Properties of estive Serders ilks in  (icp, rot, dip, drt)
    @property
    def tholder(self):
        """Tholder property getter

        Returns:
            tholder (Tholder): instance as converted from ._sad['kt']
                or None if missing.

        """
        return Tholder(sith=self._sad["kt"]) if "kt" in self._sad else None


    @property
    def keys(self):
        """Returns list of qb64 keys from ._sad['k'].
        One for each key.
        keys property getter
        """
        return self._sad.get("k")


    @property
    def verfers(self):
        """Returns list of Verfer instances as converted from ._sad['k'].
        One for each key.
        verfers property getter
        """
        keys = self._sad.get("k")
        return [Verfer(qb64=key) for key in keys] if keys is not None else None


    @property
    def ntholder(self):
        """Returns Tholder instance as converted from ._sad['nt'] or None if missing.

        """
        return Tholder(sith=self._sad["nt"]) if "nt" in self._sad else None


    @property
    def ndigs(self):
        """
        Returns:
            (list): digs
        """
        if self.vrsn.major < 2 and self.vrsn.minor < 1 and self.ilk == Ilks.vcp:
            return None

        return self._sad.get("n")

    @property
    def ndigers(self):
        """NDigers property getter

        Returns:
            ndigers (list[Diger]): instance as converted from ._sad['n'].
            One for each next key digests.
        """
        if self.vrsn.major < 2 and self.vrsn.minor < 1 and self.ilk == Ilks.vcp:
            return None

        digs = self._sad.get("n")
        return [Diger(qb64=dig) for dig in digs] if digs is not None else None


    @property
    def bner(self):  # toader
        """
        bner (Number of backer TOAD threshold of accountable duplicity property getter
        Returns:
            (Number): of ._sad["bt"] hex number str converted. Auto converts
            hex num str to int
        """
        return Number(num=self._sad["bt"]) if 'bt' in self._sad else None


    @property
    def bn(self):
        """
        bn (backer TOAD number) property getter
        Returns:
            bn (int): of .bner.num from .ked["bt"]
        """
        return self.bner.num if self.bner is not None else None


    # properties for incentive Serders like icp, dip
    @property
    def backs(self):
        """Backers property getter

        Returns:
            backs (list[str]): aids qb64 from ._sad['b'].
                           One for each backer (witness).

        """
        return self._sad.get("b")


    @property
    def berfers(self):
        """Berfers property getter
        Returns list of Verfer instances as converted from ._sad['b'].
                One for each backer (witness).

        """
        baks = self._sad.get("b")
        return [Verfer(qb64=bak) for bak in baks] if baks is not None else None


    # properties for priorative Serders like ixn rot drt

    @property
    def prior(self):
        """Prior property getter
        Returns:
            prior (str): said qb64 of prior event from ._sad['p'].

        """
        return self._sad.get("p")


    @property
    def priorb(self):
        """Priorb bytes property getter
        Returns:
            priorb (str): said qb64b of prior event from ._sad['p'].

        """
        return self.prior.encode("utf-8") if self.prior is not None else None


    # properties for rotative Serders like rot drt

    @property
    def cuts(self):
        """Cuts property getter
        Returns list of aids of instances as converted from ._sad['br'].
                 One for each backer (witness) to be cut (removed).

        """
        return self._sad.get("br")


    @property
    def adds(self):
        """Adds property getter
        Returns list of aids of instances as converted from ._sad['ba'].
                 One for each backer (witness) to be added.

        """
        return self._sad.get("ba")


    #Properties for delegated Serders ilks in (dip, drt)

    @property
    def delpre(self):
        """
        Returns:
           delpre (str): qb64  of .sad["di"] delegator ID prefix property getter
        """
        return self._sad.get("di")


    @property
    def delpreb(self):
        """
        Returns:
        delpreb (bytes): qb64b  of .delpre property getter as bytes
        """
        return self.delpre.encode("utf-8") if self.delpre is not None else None

    #Propertives for dated Serders, qry, rpy, pro, bar, exn

    @property
    def stamp(self):
        """
        Returns:
           stamp (str): date-time-stamp sad["dt"]. RFC-3339 profile of ISO-8601
                datetime of creation of message or data
        """
        return self._sad.get("dt")


    #Properties for exn  exchange


    #Properties for vcp  (registry  inception event)
    @property
    def uuid(self):
        """uuid property getter

        Returns:
           uuid (str): qb64  of .sad["u"] salty nonce
        """
        return self._sad.get("u")

    @property
    def nonce(self):
        """
        should be deprecated

        Returns:
           nonce (str): alias for .uuid property
        """
        if self.vrsn.major < 2 and self.vrsn.minor < 1 and self.ilk == Ilks.vcp:
            return self._sad.get("n")
        else:
            return self.uuid


class SerderCREL(Serder):
    """SerderCREL is Serder subclass with Labels for CREL packet types (ilks) and
       properties for exposing field values of CREL messages
       Container Registry Event Log for issuance, revocation, etc registries of
       ACDC

       See docs for Serder
    """
    #override in subclass to enforce specific protocol
    Protocol = Protos.crel  # required protocol, None means any in Protos is ok
    Proto = Protos.crel  # default protocol type
    Vrsn = Vrsn_1_1  # default protocol version for protocol type


    def _verify(self, **kwa):
        """Verifies said(s) in sad against raw
        Override for protocol and ilk specific verification behavior. Especially
        for inceptive ilks that have more than one said field like a said derived
        identifier prefix.

        Raises a ValidationError (or subclass) if any verification fails

        """
        super(SerderCREL, self)._verify(**kwa)

        try:
            code = Matter(qb64=self.issuer).code
        except Exception as ex:
            raise ValidationError(f"Invalid issuer AID = "
                                  f"{self.issuer}.") from ex

        if code not in PreDex:
            raise ValidationError(f"Invalid issuer AID code = {code}.")


    @property
    def issuer(self):
        """
        Returns:
           issuer (str): qb64  of .sad["i"] issuer AID property getter
        """
        return self._sad.get('i')


    @property
    def issuerb(self):
        """
        Returns:
        issuerb (bytes): qb64b  of .issuer property getter as bytes
        """
        return self.issuer.encode("utf-8") if self.issuer is not None else None


class SerderACDC(Serder):
    """SerderACDC is Serder subclass with Labels for ACDC packet types (ilks) and
       properties for exposing field values of ACDC messages

       See docs for Serder
    """
    #override in subclass to enforce specific protocol
    Protocol = Protos.acdc  # required protocol, None means any in Protos is ok
    Proto = Protos.acdc  # default protocol type



    def _verify(self, **kwa):
        """Verifies said(s) in sad against raw
        Override for protocol and ilk specific verification behavior. Especially
        for inceptive ilks that have more than one said field like a said derived
        identifier prefix.

        Raises a ValidationError (or subclass) if any verification fails

        """
        super(SerderACDC, self)._verify(**kwa)

        try:
            code = Matter(qb64=self.issuer).code
        except Exception as ex:
            raise ValidationError(f"Invalid issuer AID = "
                                  f"{self.issuer}.") from ex

        if code not in PreDex:
            raise ValidationError(f"Invalid issuer AID code = {code}.")

    @property
    def uuid(self):
        """uuid property getter
        Optional fields return None when not present
        Returns:
           uuid (str | None): qb64  of .sad["u"] salty nonce
        """
        return self._sad.get("u")


    @property
    def uuidb(self):
        """uuid property getter (uuid bytes)
        Optional fields return None when not present
        Returns:
           uuidb (bytes | None): qb64b  of .sad["u"] salty nonce as bytes
        """
        return self.uuid.encode("utf-8") if self.uuid is not None else None


    @property
    def issuer(self):
        """issuer property getter (issuer AID)
        Optional fields return None when not present
        Returns:
           issuer (str | None): qb64  of .sad["i"] issuer AID
        """
        return self._sad.get('i')


    @property
    def issuerb(self):
        """issuerb property getter (issuer AID bytes)
        Optional fields return None when not present
        Returns:
        issuerb (bytes | None): qb64b  of .issuer AID as bytes
        """
        return self.issuer.encode("utf-8") if self.issuer is not None else None


    @property
    def regi(self):
        """regi property getter (registry identifier SAID)
        Optional fields return None when not present
        Returns:
           regi (str | None): qb64  of .sad["ri"] registry SAID
        """
        return self._sad.get('ri')


    @property
    def regib(self):
        """regib property getter (registry identifier SAID bytes)
        Optional fields return None when not present
        Returns:
        regib (bytes | None): qb64b  of .issuer AID as bytes
        """
        return self.issuer.encode("utf-8") if self.issuer is not None else None


    @property
    def schema(self):
        """schema block or SAID property getter
        Optional fields return None when not present
        Returns:
            schema (dict | str | None): from ._sad["s"]
        """
        return self._sad.get('s')


    @property
    def attrib(self):
        """attrib block or SAID property getter (attribute)
        Optional fields return None when not present
        Returns:
            attrib (dict | str | None): from ._sad["a"]
        """
        return self._sad.get("a")


    @property
    def issuee(self):
        """ise property getter (issuee AID)
        Optional fields return None when not present
        Returns:
           issuee (str | None): qb64  of .sad["a"]["i"] issuee AID
        """
        try:
            return self.attrib.get['i']
        except:
            return None


    @property
    def issueeb(self):
        """isrb property getter (issuee AID bytes)
        Optional fields return None when not present
        Returns:
        issueeb (bytes | None): qb64b  of .issuee AID as bytes
        """
        return self.issuee.encode("utf-8") if self.issuee is not None else None


    @property
    def attagg(self):
        """Attagg block property getter (attribute aggregate)
        Optional fields return None when not present
        Returns:
            attagg (dict | str): from ._sad["A"]
        """
        return self._sad.get("A")


    @property
    def edge(self):
        """Edge block property getter
        Optional fields return None when not present
        Returns:
            edge (dict | str): from ._sad["e"]
        """
        return self._sad.get("e")


    @property
    def rule(self):
        """Rule block property getter
        Optional fields return None when not present

        Returns:
            rule (dict | str): from ._sad["r"]
        """
        return self._sad.get("r") # or {}  # need to fix logic so can remove or since optional

    # ToDo Schemer property getter. Schemer object should change name to Schemar



# -*- encoding: utf-8 -*-
"""
keri.kli.commands.oobi module

"""
import argparse

from keri import help
from hio.base import doing

import keri.app.oobiing
from keri.app import habbing, oobiing
from keri.app.cli.common import existing
from keri.db import basing
from keri.help import helping

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description="Resolve the provided OOBI")
parser.set_defaults(handler=lambda args: resolve(args),
                    transferable=True)

# Parameters for basic structure of database
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument("--oobi", "-o", help="out-of-band introduciton to load", required=True)
parser.add_argument("--oobi-alias", dest="oobiAlias", help="alias for AID resolved from out-of-band introduciton",
                    required=False, default=None)
parser.add_argument('--force', action="store_true", required=False,
                    help='True means to resolve OOBI even if it has already been previously resolved')


# Parameters for Manager access
# passcode => bran
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)


def resolve(args):
    """ command line method for resolving oobies

    Parameters:
        args(Namespace): parse args namespace object

    """
    name = args.name
    base = args.base
    bran = args.bran
    oobi = args.oobi
    oobiAlias = args.oobiAlias
    force = args.force

    icpDoer = OobiDoer(name=name, oobi=oobi, bran=bran, base=base, oobiAlias=oobiAlias, force=force)

    doers = [icpDoer]
    return doers


class OobiDoer(doing.DoDoer):
    """ DoDoer for loading oobis and waiting for the results """

    def __init__(self, name, oobi, oobiAlias, force=False, bran=None, base=None):

        self.processed = 0
        self.oobi = oobi
        self.force = force
        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.hbyDoer = habbing.HaberyDoer(habery=self.hby)

        obr = basing.OobiRecord(date=helping.nowIso8601())
        if oobiAlias is not None:
            obr.oobialias = oobiAlias

        self.hby.db.oobis.put(keys=(oobi,), val=obr)

        self.obi = keri.app.oobiing.Oobiery(hby=self.hby)
        self.authn = oobiing.Authenticator(hby=self.hby)
        doers = [self.hbyDoer, doing.doify(self.waitDo)]

        super(OobiDoer, self).__init__(doers=doers)

    def waitDo(self, tymth, tock=0.0):
        """ Waits for oobis to load

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        Returns:  doifiable Doist compatible generator method for loading oobis using
        the Oobiery
        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        if self.force:  # if --force specified, remove previous record of OOBI resolution
            self.hby.db.roobi.rem(keys=(self.oobi,))

        self.extend(self.obi.doers)
        self.extend(self.authn.doers)

        while not self.obi.hby.db.roobi.get(keys=(self.oobi,)):
            yield 0.25

        obr = self.obi.hby.db.roobi.get(keys=(self.oobi,))
        print(self.oobi, obr.state)

        self.remove([self.hbyDoer, *self.obi.doers, *self.authn.doers])
